import logging
import time
import uuid

import rest_framework_filters as filters
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import Http404, HttpResponse
from django_filters.constants import EMPTY_VALUES
from rest_framework import serializers, status
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.views import APIView
from unplugged import (
    CommandBaseMeta,
    CommandViewMixin,
    JSONAPIObject,
    JSONAPIRoot,
    JSONSchema,
    Schema,
    command,
    fields,
)
from unplugged.models import Log

from ...stream import create_stream
from .models import ListingItem

logger = logging.getLogger(__name__)


class ViewState(dict):
    """
    Information about someone consuming something,
    e.g. current subtitle or timestamp.
    """

    def __init__(self, user, id):
        if not id:
            id = str(uuid.uuid4())

        self.id = id
        self.user = user

        self.callbacks = []
        self.player_connected = False

    def add_change_callback(self, fn):
        """
        Add plugin to inform about view state changes.
        """
        self.callbacks.append(fn)

    def do_callback(self):
        logger.debug("Doing viewstate callbacks")
        if self.player_connected:
            for fn in self.callbacks:
                fn(self)

    def update(self, values):
        super(ViewState, self).update(values)
        self.do_callback()

    def __setitem__(self, key, value):
        super(ViewState, self).__setitem__(key, value)
        self.do_callback()

    def __delitem__(self, key):
        super(ViewState, self).__delitem__(key)
        self.do_callback()

    def serialize(self):
        retval = {}
        retval.update(self)
        retval["id"] = self.id
        return retval


class LimitablePageNumberPagination(PageNumberPagination):
    page_size = 100
    page_size_query_param = "limit"
    max_page_size = 1000


class CommandSerializer(serializers.Serializer):
    command = serializers.CharField()
    kwargs = serializers.DictField(required=False)


class TagSchema(Schema):
    handler = fields.String(required=True)
    tag_name = fields.String(required=True)


class DefaultFilterSet(filters.FilterSet):
    def __init__(self, data=None, *args, **kwargs):
        if data is not None and "defaults" in kwargs:
            defaults = kwargs.pop("defaults")
            # get a mutable copy of the QueryDict
            data = data.copy()

            for name, f in self.base_filters.items():
                default = defaults.get(name)

                # filter param is either missing or empty, use default
                if not data.get(name) and default:
                    data[name] = default

        super(DefaultFilterSet, self).__init__(data, *args, **kwargs)


class AggregatedOrderingFilter(filters.OrderingFilter):
    def __init__(self, *args, **kwargs):
        """
        ``field_aggregation`` Tuple of how to aggregate this field
        """
        self.field_aggregation = kwargs.pop("field_aggregation", {})
        super(AggregatedOrderingFilter, self).__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        ordering = []
        for param in value:
            descending = param.startswith("-")
            param = param[1:] if descending else param
            field_name = self.param_map.get(param, param)

            if field_name in self.field_aggregation:
                agg_name = "_agg_%s" % (field_name,)
                m = self.field_aggregation[field_name][descending]
                qs = qs.annotate(**{agg_name: m(field_name)})
                field_name = agg_name

            ordering.append("-%s" % field_name if descending else field_name)

        return qs.order_by(*ordering)


class BaseListingView(
    APIView, PermissionRequiredMixin, CommandViewMixin, metaclass=CommandBaseMeta
):
    """
    Builds a listing view and returns it. The view also handles streaming of files.
    """

    pagination_class = LimitablePageNumberPagination
    service = None
    listing_builder = None

    def get(self, request, path):
        """
        List a folder or item
        """
        path = path.strip("/")
        if path == "":
            logger.info("Root path, should just get listing configuration")
            return Response(self.get_config_view(request))

        path = self.get_path(request, path)

        config = self.service.get_path_config(path)
        logger.debug(f"Using config {config!r}")

        if config is None or "level" not in config:
            listingitem = self.listing_builder.get_item(path)
            if not listingitem:
                raise Http404

            if not listingitem.is_root:
                config, parent_path = self.service.get_nearest_config(path)
                if config["level"].get("sub_content_type"):
                    config["level"]["content_type"] = config["level"][
                        "sub_content_type"
                    ]
                config["parent_level_config"] = config["level"]
                listing = self.serialize_listing(
                    request, config, None, [listingitem], {}, {}
                )
                return Response(listing)

        do_not_rebuild = False
        current_page = request.GET.get(self.pagination_class.page_query_param, "1")
        logger.debug(
            f"Rebuild automatically status rebuild_automatically:{config.get('rebuild_automatically')} current_level:{config.get('current_level')} page:{current_page}"
        )
        if (
            current_page != "1"
            and config.get("rebuild_automatically")
            and config.get("current_level") == 0
        ):
            do_not_rebuild = True

        listing_item_root = self.listing_builder.get_listing(
            config, path, use_background_recheck=True, do_not_rebuild=do_not_rebuild
        )
        if not listing_item_root:
            raise Http404

        listingitems, links, meta = self.filter_and_paginate_listing(
            request, config, listing_item_root
        )
        listing = self.serialize_listing(
            request, config, listing_item_root, listingitems, links, meta
        )

        return Response(listing)

    def get_path(self, request, path):
        """
        Allows for internal path rewrite.
        """
        return path

    def serialize_listing(
        self, request, config, listing_item_root, listingitems, links, meta
    ):
        """
        Serializes listingitems related to root.
        """
        root = JSONAPIRoot()
        root.links.update(links)
        root.meta.update(meta)

        root.meta["content_type"] = config["level"].get("content_type", "")

        tag_info_objs = self.get_tag_info(config["level"])
        commands_obj = self.get_jsonapi_commands()

        def add_commands(obj, commands):
            for command_obj in commands:
                if command_obj._original_object.metadata.get("usable", "all") in [
                    "all",
                    obj.type,
                ]:
                    obj.add_relationship("metadata_commands", command_obj)

        if listing_item_root:
            parent = JSONAPIObject(
                listing_item_root.item_type,
                listing_item_root.get_full_path(),
                listing_item_root,
            )
            parent.update(listing_item_root.attributes)

            root.meta["parent"] = {"type": parent.type, "id": parent.id}

            root.add_included(parent)

            if config["parent_level_config"]:
                for metadata_handler in config["parent_level_config"][
                    "metadata_handlers"
                ]:
                    logger.debug(
                        f"Finding metadata from metadata handler {metadata_handler.name} for parent"
                    )
                    metadata_handler.populate_metadata_jsonapi(request, root)
                    logger.debug(
                        f"Finished metadata from metadata handler {metadata_handler.name}"
                    )

            add_commands(parent, commands_obj)

            self.add_additional_parent_relationships(
                parent, request, config, listing_item_root
            )

            for tag_info_obj in tag_info_objs:
                parent.add_relationship("metadata_tag", tag_info_obj)
        else:
            parent = None

        if config["edge_type"] == "folder":
            next_level_config = config["levels"][config["current_level"] + 1]
            tag_info_objs = self.get_tag_info(next_level_config)
        else:
            tag_info_objs = None

        for listingitem in listingitems:
            obj = JSONAPIObject(
                listingitem.item_type, listingitem.get_full_path(), listingitem
            )
            attributes = listingitem.attributes

            obj.update(attributes)
            obj["datetime"] = listingitem.datetime
            root.append(obj)

            if parent:
                obj.add_relationship("parent", parent)
            add_commands(obj, commands_obj)

            if tag_info_objs:
                for tag_info_obj in tag_info_objs:
                    obj.add_relationship("metadata_tag", tag_info_obj)

        for metadata_handler in config["level"]["metadata_handlers"]:
            logger.debug(
                f"Finding metadata from metadata handler {metadata_handler.name}"
            )
            start_time = time.time()
            metadata_handler.populate_metadata_jsonapi(request, root)
            logger.debug(
                f"Finished metadata from metadata handler {metadata_handler.name} it took {time.time() - start_time}"
            )

        if "parent" not in root.meta and root.data:
            obj = root.data[0]
            root.meta["parent"] = {"type": obj.type, "id": obj.id}

        return root.serialize(request)

    def add_additional_parent_relationships(
        self, parent, request, config, listing_item_root
    ):
        """
        If more relations needs to be added to the parent object
        """

    def serialize_filter(self, filter_, has_indexer=False):
        """
        Serializes a filter so it can be added to a JSONAPI response for the client to understand and use.
        """

        def expand_filter(filter_, choices, order_by, prefix=None):
            keys = []
            for base_name, field in filter_.base_filters.items():
                if "__" in base_name:
                    continue

                if prefix:
                    name = f"{prefix}__{base_name}"
                else:
                    name = base_name

                if isinstance(field, filters.RelatedFilter):
                    keys += expand_filter(field.filterset, choices, order_by, name)
                elif isinstance(field, filters.OrderingFilter):
                    if prefix is None:
                        order_by.extend([x[0] for x in field.extra["choices"]])
                else:
                    keys.append(name)

                if (
                    hasattr(filter_.Meta, "include_related")
                    and base_name in filter_.Meta.include_related
                ):
                    field = filter_.base_filters[base_name].field_name
                    choices[name] = [
                        x
                        for x in filter_.Meta.model.objects.distinct()
                        .values_list(field, flat=True)
                        .order_by(field)
                        if x
                    ]

            return keys

        choices = {}
        order_by = []
        fields = expand_filter(filter_, choices, order_by)
        if has_indexer:
            fields.append("q")

        return {
            "fields": fields,
            "order_by": [o for o in order_by if o],
            "choices": choices,
        }

    def get_tag_info(self, level_config):
        """
        Create tag to be added as relation to reply. Makes it possible to tag metadata.
        """
        tags = []
        for tag in level_config.get("tags", []):
            obj = JSONAPIObject("metadata_tag", tag.name)
            obj["plugin_name"] = tag.plugin_name

            tags.append(obj)

        return tags

    def get_level_filter_info(self, name, level_config):
        """
        Create filterinfo to be added as relation to reply. Makes it possible to create
        a rich search interface.
        """
        obj = JSONAPIObject("metadata_filterinfo", name)

        metadata_handlers = level_config.get("metadata_handlers", [])
        obj["metadata_handlers"] = [
            f"metadata_{mh.plugin_name}" for mh in metadata_handlers
        ]

        filter_ = self.create_filter(metadata_handlers)
        obj["filter"] = self.serialize_filter(
            filter_, has_indexer=bool(level_config.get("indexer"))
        )

        content_type = level_config.get("content_type")
        if content_type:
            obj["content_type"] = content_type

        return obj

    def get_config_view(self, request):
        """
        Returns the configuration for this listingview. Can be used by clients to setup the correct
        sections etc.
        """
        raise NotImplementedError()

    def create_filter(self, metadata_handlers):
        """
        Creates the filter that can be used to filter the queryset.
        """
        f = {}
        order_by_mapping = {"path": "path", "datetime": "datetime"}
        order_by_aggregate = {}
        user_fields = []
        for metadata_handler in metadata_handlers:
            logger.debug("Checking %s for filter" % metadata_handler.name)
            if hasattr(metadata_handler, "filter") and metadata_handler.filter:
                name = f"metadata_{metadata_handler.plugin_name}"
                filter_name = f"metadata_{metadata_handler.plugin_name}"
                logger.debug(f"Adding {name} to filter")
                f[name] = filters.RelatedFilter(
                    metadata_handler.filter,
                    queryset=metadata_handler.listing_item_relation_model.objects.all(),
                )

                filter_order_by_aggregate = getattr(
                    metadata_handler.filter.Meta, "order_by_aggregate", {}
                )
                if hasattr(metadata_handler.filter.Meta, "order_by"):
                    for key in metadata_handler.filter.Meta.order_by or []:
                        logger.debug(f"Found ordering key {key}")
                        field = metadata_handler.filter.base_filters[key]
                        aggregators = filter_order_by_aggregate.get(key)

                        key = "%s__%s" % (name, key)
                        target = "%s__%s" % (filter_name, field.field_name)
                        order_by_mapping[target] = key
                        if aggregators:
                            order_by_aggregate[target] = aggregators

                if hasattr(metadata_handler.filter.Meta, "user_field"):
                    user_fields.append(
                        (
                            "%s__" % (name,),
                            "%s__%s" % (name, metadata_handler.filter.Meta.user_field),
                        )
                    )

        logger.info(f"SectionsFilter built with {len(f)} keys")

        f["o"] = AggregatedOrderingFilter(
            fields=order_by_mapping.items(), field_aggregation=order_by_aggregate
        )

        class Meta:
            model = ListingItem
            fields = ["path", "datetime"]

        Meta.user_fields = user_fields
        f["Meta"] = Meta

        filter_ = type("MetadataFilter", (DefaultFilterSet,), f)

        return filter_

    def filter_and_paginate_listing(self, request, config, listing_item_root):
        qs = listing_item_root.listingitem_set.all().distinct()

        indexer = config["level"].get("indexer")
        if indexer and "q" in request.GET:
            paths = indexer.search(listing_item_root.path, request.GET["q"])
            qs = qs.filter(path__in=paths)

        filter_ = self.create_filter(config["level"]["metadata_handlers"])
        ordering = request.GET.get("o") or config["level"].get("default_ordering", "")
        listingitems = filter_(request.GET, queryset=qs, defaults={"o": ordering}).qs
        for k in list(request.GET.keys()) + ordering.split(","):
            for prefix, field in filter_.Meta.user_fields:
                if k.startswith(prefix):
                    listingitems = listingitems.filter(**{field: request.user})
        page, listingitems = self.paginate_queryset(listingitems)

        seen = set()
        seen_add = seen.add
        listingitems = [x for x in listingitems if not (x in seen or seen_add(x))]

        page_links = {}
        next_page = page.get_next_link()
        if next_page:
            logger.debug("Found next page")
            page_links["next"] = next_page

        prev_page = page.get_previous_link()
        if prev_page:
            logger.debug("Found previous page")
            page_links["prev"] = prev_page

        meta = {
            "count": page.page.paginator.count,
            "num_pages": page.page.paginator.num_pages,
            "page": page.page.number,
            "limit": page.page.paginator.per_page,
        }

        return listingitems, page_links, meta

    @command("stream", metadata={"usable": "file"}, display_name="Stream")
    def stream(self, request, config, listingitem):
        item = listingitem.get_original_item()

        streamresult = create_stream(item, request)

        # UserActionLog.objects.log_plugin(self.service, request.user, 'Stream started by %s' % (request.user, ),
        #                                  'Stream for %s started by %s' % (item.path, request.user))
        with Log.objects.start_chain(
            self.service, "USER.STREAM_START", user=request.user
        ) as log:
            log.log(0, f"Trying to stream {item.path}")

            viewstate = ViewState(request.user, request.GET.get("viewstate"))
            root = JSONAPIRoot()

            for history_plugin in config.get("histories", []):
                logger.debug(f"Logging history to {history_plugin.name}")
                history_plugin.log_history(config, listingitem, viewstate)

            viewstate_obj = JSONAPIObject("viewstate", viewstate.id)
            viewstate_obj.update(viewstate)
            root.append(viewstate_obj)

            player_service = self.service.config.get("player_service")
            player_id = self.request.GET.get("target")
            if player_id and player_service:
                payload = streamresult.serialize(request)
                player_service.play(payload, viewstate, player_id)

            root.append(streamresult.to_jsonapi(request))

            return Response(root.serialize(request))

    def get_tag_plugin(self, config, handler):
        tag_handlers = [x for x in config["level"].get("tags", []) if x.name == handler]
        if not tag_handlers:
            return None

        return tag_handlers[0]

    @command(schema=TagSchema)
    def command_tag(self, request, config, listingitem, handler, tag_name):
        tag_plugin = self.get_tag_plugin(config, handler)
        if not tag_plugin:
            jsonapi_root = JSONAPIRoot.error_status(
                id_="invalid_tag", detail=f"Tag {handler!r} is not a known tag"
            )
            return Response(
                jsonapi_root.serialize(request), status=status.HTTP_400_BAD_REQUEST
            )

        tag_plugin.tag_listingitem(config, request.user, listingitem, tag_name)
        return Response(
            JSONAPIRoot.success_status("Item tagged successfully").serialize(request)
        )

    @command(schema=TagSchema)
    def command_untag(self, request, config, listingitem, handler, tag_name):
        tag_plugin = self.get_tag_plugin(config, handler)
        if not tag_plugin:
            jsonapi_root = JSONAPIRoot.error_status(
                id_="invalid_tag", detail=f"Tag {handler!r} is not a known tag"
            )
            return Response(
                jsonapi_root.serialize(request), status=status.HTTP_400_BAD_REQUEST
            )

        tag_plugin.untag_listingitem(config, request.user, listingitem, tag_name)
        return Response(
            JSONAPIRoot.success_status("Item untagged successfully").serialize(request)
        )

    def post(self, request, path):
        """
        Do stuff to an item.
        """
        path = path.strip("/")
        path = self.get_path(request, path)

        config, parent_path = self.service.get_nearest_config(path)

        logger.info(f"Trying to get listingitem for path {path!r}")

        if config is None:
            raise Http404

        try:
            listingitem = ListingItem.objects.get(app=self.service.name, path=path)
        except ListingItem.DoesNotExist:
            raise Http404

        kwargs = {"request": request, "config": config, "listingitem": listingitem}

        return self.call_command(request, self, kwargs)

    @property
    def paginator(self):
        """
        The paginator instance associated with the view, or `None`.
        """
        if not hasattr(self, "_paginator"):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_queryset(self, queryset):
        """
        Return a single page of results, or `None` if pagination is disabled.
        """
        if self.paginator is None:
            return None
        return (
            self.paginator,
            self.paginator.paginate_queryset(queryset, self.request, view=self),
        )



import os

from django.conf import settings
from django.conf.urls import url
from unplugged import RelatedPluginField, Schema, ServicePlugin, fields

from ...plugins import ConfigPlugin
from ...utils import serve
from .views import IndexView


class WebinterfaceSchema(Schema):
    mount_at_root = fields.Boolean()
    config = RelatedPluginField(plugin_type=ConfigPlugin)


class FrontpageConfigSourceQuerySchema(Schema):
    key = fields.String(required=True)
    value = fields.String(required=True)


class FrontpageConfigDocChainSchema(Schema):
    type = fields.String(required=True)
    id = fields.String(required=False)


class FrontpageConfigSourceSchema(Schema):
    docChain = fields.Nested(FrontpageConfigDocChainSchema, many=True, required=True)
    contentType = fields.String(
        enum=settings.TRIDENTSTREAM_CONTENT_TYPES,
        enum_names=settings.TRIDENTSTREAM_CONTENT_TYPES_NAMES,
        default="",
    )
    template = fields.String(
        enum=settings.TRIDENTSTREAM_TEMPLATES,
        enum_names=settings.TRIDENTSTREAM_TEMPLATES_NAMES,
        default="cover",
        required=True,
    )
    query = fields.Nested(FrontpageConfigSourceQuerySchema, many=True)
    enabled = fields.Boolean()


class FrontpageConfigSchema(Schema):
    title = fields.String(required=True)
    sources = fields.Nested(FrontpageConfigSourceSchema, many=True, required=True)
    enabled = fields.Boolean()


class FrontpageConfigRootSchema(Schema):
    sections = fields.Nested(FrontpageConfigSchema, many=True)


class WebinterfaceServicePlugin(ServicePlugin):
    plugin_name = "webinterface"

    config_schema = WebinterfaceSchema

    default_config = {
        "sections": [
            {
                "title": "Shows you follow",
                "sources": [
                    {
                        "docChain": [{"type": "service_sections"}, {"type": "folder"}],
                        "contentType": "tvshows",
                        "template": "cover",
                        "enabled": True,
                        "query": [
                            {"key": "limit", "value": "12"},
                            {"key": "o", "value": "-datetime"},
                            {"key": "metadata_tag__tag_name", "value": "follow"},
                        ],
                    }
                ],
                "enabled": True,
            },
            {
                "title": "Good, new movies",
                "sources": [
                    {
                        "docChain": [{"type": "service_sections"}, {"type": "folder"}],
                        "contentType": "movies",
                        "template": "cover",
                        "enabled": True,
                        "query": [
                            {"key": "limit", "value": "24"},
                            {
                                "key": "o",
                                "value": "-metadata_firstseen__datetime,-datetime",
                            },
                        ],
                    }
                ],
                "enabled": True,
            },
        ]
    }

    def __init__(self, config):
        if config["mount_at_root"]:
            self.mount_at_root = True

        config_plugin = config.get("config")
        if config_plugin:
            namespace = f"{self.plugin_type}:{self.name}"
            key = "frontpage_listings"
            config_plugin.set_default_config(
                "system", namespace, key, self.default_config
            )
            config_plugin.set_config_schema(namespace, key, FrontpageConfigRootSchema)

    def get_urls(self):
        return [
            url(r"^/?$", IndexView.as_view(service=self)),
            url(r"^/index.html$", IndexView.as_view(service=self)),
            url(
                r"^(?P<path>.*)$",
                serve,
                kwargs={
                    "document_root": os.path.join(
                        os.path.dirname(os.path.abspath(__file__)), "static"
                    )
                },
            ),
        ]



import hashlib
import json
import logging
from abc import abstractmethod
from datetime import timedelta

from django.conf import settings
from django.core.cache import cache
from thomas import router
from unplugged import PluginBase, pluginhandler

from ..exceptions import NotModifiedException, PathNotFoundException
from ..stream import Stream
from ..utils import hash_string

logger = logging.getLogger(__name__)

LISTING_CACHE_TIME = timedelta(days=7)
THOMAS_LIST_HANDLER_NAME = "ts_list"


class InputPlugin(PluginBase):
    plugin_type = "input"

    @abstractmethod
    def get_item(self, path):
        """
        Return an item for a given path.

        Does not have to be populated with correct data but it should
        contain the correct routes.
        """


def _listing_cache_key(prefix, plugin, path, depth):
    return f"listing:{prefix}:{plugin}:{depth}:{hash_string(path)}"


class InputPluginManager:
    @staticmethod
    def get_item_multiple(
        plugin_path_pairs,
    ):  # get an item for all paths and merge into one
        item = None
        for plugin, path in plugin_path_pairs:
            plugin_item = plugin.get_item(path)
            if item:
                item.merge(plugin_item)
            else:
                item = plugin_item

        return item

    @staticmethod
    def get_item(plugin, path):
        return InputPluginManager.get_item_multiple([(plugin, path)])

    @staticmethod
    def cached_list(f, item, _route_name, **kwargs):
        key = json.dumps((item.id, _route_name, sorted(kwargs.items()))).encode("utf-8")
        cache_key = hashlib.sha1(key).hexdigest()
        cache_key_date = f"date:{cache_key}"
        cache_key_listing = f"listing:{cache_key}"

        cached_date = cache.get(cache_key_date)
        if cached_date:
            logger.debug(
                f"We found a cache date {cached_date}, looking for listing too"
            )
            cached_listing = cache.get(cache_key_listing)
        else:
            cached_listing = None

        if not cached_listing:
            cached_date = None

        listed = False
        kwargs["modified_since"] = cached_date
        try:
            listing = f(item, **kwargs)
            listed = True
        except NotModifiedException:
            listing = item.__class__.unserialize(cached_listing, router=item.router)
        except PathNotFoundException:
            logger.debug(f"Path not found on {f!r}")
            item.expandable = False
            return None
        except Exception:
            logger.exception(f"Failed to list {f!r}")
            item.expandable = False
            return None

        if listed and listing:
            cache.set(
                cache_key_date,
                listing.modified + timedelta(seconds=2),
                LISTING_CACHE_TIME.total_seconds(),
            )
            cache.set(
                cache_key_listing,
                listing.serialize(include_routes=True),
                LISTING_CACHE_TIME.total_seconds(),
            )
            logger.debug(
                f"Caching listing for item:{item} with modified:{listing.modified} and using f:{f!r} kwargs:{kwargs!r}"
            )

        return listing


router.list_decorator = InputPluginManager.cached_list



import hashlib
import os
import posixpath
import time
from urllib.parse import urljoin as _urljoin

from django.http import Http404, HttpResponse, HttpResponseNotModified
from django.utils._os import safe_join
from django.utils.six.moves.urllib.parse import unquote
from django.views.static import directory_index, was_modified_since


def urljoin(*args):
    args = list(args)
    url = args.pop(0)
    for arg in args:
        url = _urljoin(url.rstrip("/") + "/", arg.lstrip("/"))

    return url


def serve(request, path, document_root=None, show_indexes=False):
    """
    Note: Modified to use X-Sendfile
    Serve static files below a given point in the directory structure.
    To use, put a URL pattern such as::
        from django.views.static import serve
        url(r'^(?P<path>.*)$', serve, {'document_root': '/path/to/my/files/'})
    in your URLconf. You must provide the ``document_root`` param. You may
    also set ``show_indexes`` to ``True`` if you'd like to serve a basic index
    of the directory.  This index view will use the template hardcoded below,
    but if you'd like to override it, you can create a template called
    ``static/directory_index.html``.
    """
    path = posixpath.normpath(unquote(path)).lstrip("/")
    fullpath = safe_join(document_root, path)
    if os.path.isdir(fullpath):
        if show_indexes:
            return directory_index(path, fullpath)
        raise Http404("Directory indexes are not allowed here.")
    if not os.path.exists(fullpath):
        raise Http404('"%(path)s" does not exist' % {"path": fullpath})
    # Respect the If-Modified-Since header.
    statobj = os.stat(fullpath)
    if not was_modified_since(
        request.META.get("HTTP_IF_MODIFIED_SINCE"), statobj.st_mtime, statobj.st_size
    ):
        return HttpResponseNotModified()
    response = HttpResponse()
    response["X-Sendfile"] = fullpath

    return response


def hash_string(s):
    try:
        return hashlib.sha1(s).hexdigest()
    except (UnicodeEncodeError, TypeError):
        return hashlib.sha1(s.encode("utf-8")).hexdigest()



import logging

from django.apps import AppConfig as DjangoAppConfig

logger = logging.getLogger(__name__)


class AppConfig(DjangoAppConfig):
    name = "tridentstream.dbs.shelve"
    verbose_name = "Shelve Database Plugin"
    label = "db_shelve"

    def ready(self):
        from .handler import ShelfDatabasePlugin  # NOQA



from __future__ import absolute_import

import os

import environ

"""
Tridentstream Django settings.
"""

env = environ.Env(
    ALLOWED_HOSTS=(list, ["*"]),
    MEDIA_ROOT=(str, "media"),
    DATABASE_ROOT=(str, "dbs"),
    INSTALLED_APPS=(list, []),
    DISABLE_CSRF_SECURITY=(bool, True),
)
env.read_env(env.str("ENV_PATH", ".environ"))

if os.environ.get("DJANGO_DEBUG") == "1":
    LOGGING = {
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": {
            "console": {"class": "logging.StreamHandler"},
            "null": {"level": "DEBUG", "class": "logging.NullHandler"},
        },
        "loggers": {
            "django": {"handlers": ["console"], "level": "DEBUG"},
            "django.db.backends": {
                "handlers": ["null"],  # Quiet by default!
                "propagate": False,
                "level": "DEBUG",
            },
            "rebulk.rules": {
                "handlers": ["null"],  # Quiet by default!
                "propagate": False,
                "level": "DEBUG",
            },
        },
    }
    DEBUG = True
else:
    DEBUG = False

# Tridentstream specific apps

DATABASE_APPS = (
    "tridentstream.dbs.leveldb",
    # "tridentstream.dbs.lmdb", # TODO: make auto growing before re-enabling
    "tridentstream.dbs.memory",
    "tridentstream.dbs.shelve",
)

METADATA_APPS = (
    "tridentstream.metadata.imdb",
    "tridentstream.metadata.mal",
    "tridentstream.metadata.embedded",
    "tridentstream.metadata.firstseen",
    "tridentstream.metadata.history",
    "tridentstream.metadata.iteminfo",
    "tridentstream.metadata.thetvdb",
    "tridentstream.metadata.tag",
    "tridentstream.metadata.name",
    "tridentstream.metadata.available",
)

INPUT_APPS = ("tridentstream.inputs.fs", "tridentstream.inputs.rfs")

BASE_APPS = (
    "tridentstream.bases.listing",
    "tridentstream.bases.metadata",
    "tridentstream.bases.websearcher",
)

INDEXER_APPS = ("tridentstream.indexers.whoosh",)

SERVICE_APPS = (
    "tridentstream.services.imgcache",
    "tridentstream.services.metadata",
    "tridentstream.services.config",
    "tridentstream.services.remotesearcher",
    "tridentstream.services.rfs",
    "tridentstream.services.sections",
    "tridentstream.services.store",
    "tridentstream.services.player",
    "tridentstream.services.webinterface",
)

BITTORRENT_CLIENT_APPS = ("tridentstream.bittorrentclients.deluge",)

SEARCHER_APPS = ("tridentstream.searchers.remotesearcher",)

NOTIFIER_APPS = (
    "tridentstream.notifiers.wamp",
    "tridentstream.notifiers.multinotifier",
)

# Application definition

INSTALLED_APPS = (
    (
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "rest_framework",
        "rest_framework.authtoken",
        "rest_framework_filters",
        "corsheaders",
        "channels",
        "unplugged",
        "unplugged.services.admin",
        "unplugged.services.api",
        "unplugged.services.staticurls",
        "unplugged.services.user",
        "unplugged.services.wamp",
        "tridentstream",
    )
    + BASE_APPS
    + DATABASE_APPS
    + INPUT_APPS
    + METADATA_APPS
    + SERVICE_APPS
    + INDEXER_APPS
    + SEARCHER_APPS
    + BITTORRENT_CLIENT_APPS
    + NOTIFIER_APPS
)


MIDDLEWARE = [
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]


ROOT_URLCONF = "main.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    }
]

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [  # TODO: feedback needed in webinterface
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# DRF settings

REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 50,
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.BasicAuthentication",
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.TokenAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.IsAuthenticated",),
}

APPEND_SLASH = False
# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = "/djangostatic/"
MEDIA_URL = "/media/"

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

DEBUG_PROPAGATE_EXCEPTIONS = False

FILESERVE_RESOURCE = None

CORS_ORIGIN_ALLOW_ALL = True
CORS_ALLOW_CREDENTIALS = True

DEBUG_CLEAR_CACHE = False

PLUGIN_INITIALIZATION_ORDER = ["service"]

TRIDENTSTREAM_TEMPLATES = ("cover", "table")

TRIDENTSTREAM_TEMPLATES_NAMES = ("Cover", "Table")

TRIDENTSTREAM_CONTENT_TYPES = (
    "songs",
    "albums",
    "movies",
    "movie",
    "tvshows",
    "seasons",
    "episodes",
)

TRIDENTSTREAM_CONTENT_TYPES_NAMES = (
    "Songs",
    "Albums",
    "Movies",
    "Movie",
    "TV Shows",
    "Seasons",
    "Episodes",
)

ASGI_APPLICATION = "main.routing.application"
SCHEDULER = None

THOMAS_STREAMER_PLUGINS = ["direct", "rar"]

VIDEO_STREAMABLE_EXTENSIONS = ["mkv", "mp4", "iso", "ogg", "ogm", "m4v", "wmv"]
AUDIO_STREAMABLE_EXTENSIONS = ["flac", "mp3", "oga", "m4a"]
STREAMABLE_EXTENSIONS = set(VIDEO_STREAMABLE_EXTENSIONS + AUDIO_STREAMABLE_EXTENSIONS)
WAMP_REALM = "tridentstream"

CACHES = {
    "default": env.cache(default="dbcache://cache_table"),
}

DATABASES = {
    "default": env.db(default="sqlite:///db.sqlite3?timeout=120"),
}

INSTALLED_APPS += tuple(env("INSTALLED_APPS"))

SECRET_KEY = env("SECRET_KEY")

ALLOWED_HOSTS = env("ALLOWED_HOSTS")

MEDIA_ROOT = env("MEDIA_ROOT")
DATABASE_ROOT = env("MEDIA_ROOT")

if env("DISABLE_CSRF_SECURITY"):
    MIDDLEWARE += ("main.disablecsrf.DisableCSRFMiddleware",)

for option, option_value in DATABASES["default"].get("OPTIONS", {}).items():
    if option == "timeout":
        DATABASES["default"]["OPTIONS"][option] = int(option_value)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.sections"
    verbose_name = "Sections Service"
    label = "services_sections"

    def ready(self):
        from .handler import SectionsService  # NOQA



import logging
from datetime import timedelta
from threading import Event, Lock

from django.utils.timezone import now

logger = logging.getLogger(__name__)


class PathLockTracker:
    def __init__(self, locktracker, path):
        self.path = path
        self.locktracker = locktracker
        self.waited = False
        self.expired = False

    def __enter__(self):
        logger.debug(f"Trying to figure out how to handle the lock of path:{self.path}")

        found_existing_lock = False
        with self.locktracker.build_lock:
            if self.path in self.locktracker.locks:
                logger.debug("Existing lock found, lets check if it is timed out yet")
                event, started = self.locktracker.locks[self.path]
                end_time = started + self.locktracker.wait_time
                if end_time < now():
                    logger.info("Found lock that is timed out.")
                else:
                    found_existing_lock = True

            if not found_existing_lock:
                logger.debug("No locks found, lets create one")
                self.locktracker.locks[self.path] = Event(), now()

        if found_existing_lock:
            logger.debug("Found working existing lock, waiting for that.")
            self.waited = True
            wait_time = (now() - end_time) + timedelta(seconds=1)
            if not event.wait(wait_time.total_seconds()):
                self.expired = True

        return self

    def __exit__(self, type_, value, traceback):
        logger.debug(f"Done with one event of path:{self.path}")
        if not self.waited:
            with self.locktracker.build_lock:
                if self.path in self.locktracker.locks:
                    del self.locktracker.locks[self.path]
                else:
                    logger.warning(f"Was unable to delete event for path:{self.path}")


class LockTracker:
    wait_time = timedelta(minutes=5)

    def __init__(self):
        self.build_lock = Lock()
        self.locks = {}

    def get_path(self, path):
        logger.debug(f"Handling lock for path:{path}")
        return PathLockTracker(self, path)



import logging

from django.http import Http404
from rest_framework import permissions
from rest_framework.response import Response
from rest_framework.views import APIView
from unplugged import JSONAPIObject, JSONAPIRoot, JSONSchema

logger = logging.getLogger(__name__)


class ConfigView(APIView):
    service = None
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request):
        root = JSONAPIRoot()

        namespace = request.GET.get("namespace")
        key = request.GET.get("key")
        if not key or not namespace:
            raise Http404

        identifier = "%s/%s" % (namespace, key)

        obj = JSONAPIObject("setting", identifier)
        obj["namespace"] = namespace
        obj["key"] = key
        if request.GET.get("default"):
            obj["value"] = self.service.config.get_default_config(namespace, key)
        else:
            obj["value"] = self.service.config.get_user_config(
                request.user, namespace, key
            )

        root.append(obj)

        schema = self.service.config.get_config_schema(namespace, key)
        if schema:
            schema_obj = JSONAPIObject("schema", identifier)
            schema_obj["namespace"] = namespace
            schema_obj["key"] = key

            json_schema = JSONSchema()
            schema = schema()
            schema_obj["schema"] = json_schema.dump(schema)
            obj.add_relationship("schema", schema_obj)

        return Response(root.serialize(request))

    def delete(self, request):
        namespace = request.GET.get("namespace")
        key = request.GET.get("key")

        if request.GET.get("set_default") and request.user.is_superuser:
            self.service.config.set_default_config("configured", namespace, key, None)
        else:
            self.service.config.set_user_config(request.user, namespace, key, None)

        return Response({})

    def post(self, request):
        namespace = request.GET.get("namespace")
        key = request.GET.get("key")

        if request.GET.get("default") and request.user.is_superuser:
            self.service.config.set_default_config(
                "configured", namespace, key, request.data
            )
        else:
            self.service.config.set_user_config(
                request.user, namespace, key, request.data
            )

        return self.get(request)



import uuid

from django.conf import settings
from django.db import models
from jsonfield import JSONField

from ...bases.metadata.models import BaseListingItemRelation, BaseMetadataLink


class History(models.Model):
    app = models.CharField(max_length=30, db_index=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, db_index=True, on_delete=models.deletion.CASCADE
    )
    name = models.CharField(max_length=200, db_index=True)
    listingitem_app = models.CharField(max_length=30, null=True)
    listingitem_path = models.CharField(max_length=500, null=True)
    last_watched = models.DateTimeField()
    created = models.DateTimeField(auto_now_add=True)

    @property
    def identifier(self):
        return self.pk

    class Meta:
        ordering = ["-last_watched"]

    def __str__(self):
        return f"{self.user!r} watched {self.name} at {self.last_watched}"


class HistoryMetadata(BaseMetadataLink):
    """Metadata related to what the user saw"""

    history = models.ForeignKey(History, on_delete=models.deletion.CASCADE)
    primary_metadata = models.BooleanField(default=False)

    class Meta:
        unique_together = (("content_type", "object_id", "history"),)


class ViewState(models.Model):
    identifier = models.UUIDField(default=uuid.uuid4, editable=False)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.deletion.CASCADE
    )
    history = models.ForeignKey(History, on_delete=models.deletion.CASCADE)
    values = JSONField(default=dict, blank=True)

    last_update = models.DateTimeField(auto_now=True)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (("identifier", "user"),)


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(
        HistoryMetadata, db_index=True, on_delete=models.deletion.CASCADE
    )



import glob
import logging
import os
import time
from datetime import datetime
from queue import Queue
from urllib.parse import urljoin

import pytz
from django.conf import settings
from thomas import router
from unplugged import RelatedPluginField, Schema, command, fields, threadify
from unplugged.models import Log

from twisted.internet import defer, reactor

from ...exceptions import NotModifiedException, PathNotFoundException
from ...plugins import (
    DatabasePlugin,
    InputPlugin,
    MetadataParserPlugin,
    Notification,
    NotifierPlugin,
)
from ...vfs import FileSystem

COMMIT_COUNTER = 10000

logger = logging.getLogger(__name__)


class PathSchema(Schema):
    path = fields.String(required=True)
    virtual_root = fields.String(default="")


class FilesystemInputSchema(Schema):
    priority = fields.Integer(default=10)
    db = RelatedPluginField(plugin_type=DatabasePlugin, required=True)
    metadata_parsers = fields.List(
        RelatedPluginField(plugin_type=MetadataParserPlugin), many=True, default=list
    )
    paths = fields.Nested(PathSchema, many=True, default=list)
    notifier = RelatedPluginField(plugin_type=NotifierPlugin)


class FilesystemInputPlugin(InputPlugin):
    plugin_name = "filesystem"

    config_schema = FilesystemInputSchema

    is_rescanning = False
    rescan_done = None
    should_die = False

    vfs = None
    route_input_fs_list = None

    simpleadmin_templates = [
        {
            "template": {
                "paths": {"simpleAdminMethod": "userInput"},
                "metadata_parsers": [
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "metadataparser",
                        "name": "imdb",
                    },
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "metadataparser",
                        "name": "mal",
                    },
                ],
                "db": {
                    "simpleAdminMethod": "createPluginPriority",
                    "pluginPriorities": [
                        {
                            "plugin_type": "database",
                            "plugin_name": "leveldb",
                            "name": "db_automanaged_%(name)s",
                            "config": {"path": "db_automanaged_%(name)s"},
                        },
                        {
                            "plugin_type": "database",
                            "plugin_name": "shelf",
                            "name": "db_automanaged_%(name)s",
                            "config": {"path": "db_automanaged_%(name)s"},
                        },
                    ],
                },
                "notifier": {
                    "simpleAdminMethod": "lookupPlugin",
                    "plugin_type": "notifier",
                    "name": "multinotifier",
                },
            },
            "description": "Get data from local filesystem",
            "id": "default",
            "update_method": "full",
        }
    ]

    def __init__(self, config):
        self.db = config["db"]
        self.priority = config.get("priority", 10)
        self.vfs = FileSystem(self.db)
        self.last_update = datetime.now()
        self.vfs.add_event("new", self._new_item)
        self.vfs.add_event("file_route_needed", self._file_route_needed)
        self.vfs.add_event("list_route_needed", self._list_route_needed)
        self.vfs.add_event("on_item", self._on_item)
        self.metadata_parsers = config.get("metadata_parsers", [])
        self.notifier = config.get("notifier")

        self.paths = []
        for path_template in config.get("paths", []):
            for path in glob.glob(path_template["path"]):
                self.paths.append((path_template.get("virtual_root", ""), path))
                logger.info(
                    f"Added path {path!r} to virtual path {path_template.get('virtual_root', '')!r}"
                )

        self.route_input_fs_list = f"input_fs_list_{self.name}"

        router.register_handler(
            self.route_input_fs_list, self.thomas_list, False, True, False
        )

    @command(
        name="clear_and_rescan",
        display_name="Clear and Rescan",
        description="Clear all the stored metadata and add it anew",
    )
    def command_clear_and_rescan(self):
        logger.info("Clearing database")
        self.rescan(True)

    @command(
        name="rescan",
        display_name="Rescan",
        description="Scan for new data on the disks",
    )
    def command_rescan(self):
        logger.info("External call for rescan")
        self.rescan()

    def unload(self):
        logger.info("Asked to unload FS input handler")
        self.should_die = True
        if self.is_rescanning:
            logger.info("Unloading after rescan is done")
            self.rescan_done.addCallback(lambda x: self.close())
        else:
            logger.info("Unloading now")
            self.close()

        if self.route_input_fs_list:
            router.unregister_handler(self.route_input_fs_list)

    def _file_route_needed(self, item, path):
        item.add_route(
            "file",
            True,
            False,
            False,
            kwargs={"path": item["_actual_path"]},
            priority=self.priority,
        )

    def _list_route_needed(self, item, path):
        item.add_route(
            self.route_input_fs_list,
            False,
            True,
            False,
            kwargs={"path": path},
            priority=self.priority,
        )

    def _on_item(self, item, path):
        if (
            not item.is_listable
            and os.path.splitext(item.id)[1].lstrip(".").lower()
            not in settings.STREAMABLE_EXTENSIONS
        ):
            return

        item.streamable = True
        for streamer_plugin in settings.THOMAS_STREAMER_PLUGINS:
            item.add_route(streamer_plugin, False, False, True, priority=self.priority)

    def _new_item(self, key, db_type, metadata):
        filename = metadata["path"].split("/")[-1]
        # logger.debug('New item %s, trying to match with metadata' % (metadata['path'], ))
        for plugin in self.metadata_parsers:
            if plugin.pattern.match(filename):
                # logger.debug('Sending %s to metadataparser %s' % (metadata['path'], plugin.name))
                try:
                    plugin.handle(self.vfs, metadata["path"], metadata["_actual_path"])
                except:
                    logger.exception(
                        f"Failed to handle {metadata['_actual_path']} with plugin {plugin!r}"
                    )

    def close(self):
        if self.vfs:
            self.vfs.close()

    def list(self, path, depth, modified_since=None):
        last_modified = datetime.fromtimestamp(self.vfs.last_modified(path), pytz.UTC)

        if modified_since and last_modified <= modified_since:
            raise NotModifiedException()

        logger.info(f"Listing path {path!r} with depth {depth}")
        listing = self.vfs.list_dir(path, depth)
        return listing

    def stream(self, path):
        logger.info(f"Trying to stream {path!r}")

        item = self.list(path, -1)
        return item.stream()

    def get_item(self, path):
        try:
            return self.list(path, -1)
        except PathNotFoundException:
            logger.info(f"Getting file at path {path!r}")
            return self.vfs.list_file(path)

    def rescan(self, update_all_metadata=False):
        threadify(self._rescan)(update_all_metadata)
        return "Rescanning"

    def _rescan(self, update_all_metadata=False):
        """
        Rescans filesystem for files
        """
        if self.is_rescanning:
            logger.warning("Already rescanning")
            return None

        notification = Notification(
            f"admin.{self.plugin_name}.{self.name}.rescan",
            "info",
            f"Filesystem {self.name}",
            "Started to rescan",
            permission_type="is_admin",
        )
        notification_start_dt = datetime.now()
        if self.notifier:
            self.notifier.notify(notification)

        with Log.objects.start_chain(self, "INPUT.RESCAN") as log:
            log.log(
                0, f"A rescan is started and update_all_metadata {update_all_metadata}"
            )

            self.is_rescanning = True
            self.rescan_done = defer.Deferred()

            logger.info("Rescanning")
            queue = Queue(20)

            class QueueCommand:
                INSERT = 0
                ENSURE_PREFIXES = 1
                DONE = 2

            def walk_path(queue, prefix, path):
                logger.info(f"Starting to scan {path!r} with prefix {prefix!r}")
                list_queue = []
                queue_size = 0

                for r, dirs, files in os.walk(path, followlinks=True):
                    useful_files = []
                    for f in files:
                        full_path = os.path.join(r, f)
                        if not os.path.exists(full_path):
                            logger.debug(f"We found broken link: {full_path}")
                            continue

                        useful_files.append((f, os.path.getsize(full_path)))

                    list_queue.append((r[len(path) :].strip("/"), dirs, useful_files))
                    queue_size += len(dirs) + len(files)

                    if queue_size >= COMMIT_COUNTER:
                        queue.put((QueueCommand.INSERT, path, prefix, list_queue))
                        list_queue = []
                        queue_size = 0

                    if self.should_die:
                        logger.info(f"Got the death in walker for {path}")
                        return

                if list_queue:
                    queue.put((QueueCommand.INSERT, path, prefix, list_queue))

                queue.put((QueueCommand.DONE, path))

                logger.info(f"Done scanning {path!r}")

            def insert_into_vfs(vfs, queue, path_count):
                with vfs.session(True, always_trigger_new=update_all_metadata):
                    while path_count:
                        job = queue.get(True)
                        cmd = job[0]

                        if cmd == QueueCommand.INSERT:
                            logger.debug("Got insertion job")
                            _, path, prefix, items = job

                            for root, folders, files in items:
                                virtual_path = [prefix] + root.split(os.sep)

                                for item in folders:
                                    vp = "/".join(
                                        [x for x in virtual_path + [item] if x]
                                    )
                                    vfs.add_dir(vp, int(time.time()))

                                for item, size in files:
                                    vp = "/".join(
                                        [x for x in virtual_path + [item] if x]
                                    )
                                    actual_path = os.path.join(path, root, item)

                                    vfs.add_file(
                                        vp,
                                        size,
                                        int(time.time()),
                                        {"_actual_path": actual_path},
                                    )

                        elif cmd == QueueCommand.ENSURE_PREFIXES:
                            _, prefixes = job

                            for prefix in prefixes:
                                path = []
                                for p in prefix.strip("/").split("/"):
                                    path.append(p)
                                    vfs.add_dir("/".join(path), int(time.time()))
                        elif cmd == QueueCommand.DONE:
                            _, path = job
                            logger.info(f"Done inserting {path!r}")
                            path_count -= 1
                        else:
                            logger.error(f"Unknown command {job!r}")

                        if self.should_die:
                            logger.info(f"Got the death in inserter for {path}")
                            break

                self.last_update = datetime.now()
                self.is_rescanning = False
                reactor.callFromThread(self.rescan_done.callback, None)
                self.rescan_done.callback = None

                logger.info("Done scanning all paths.")

            t = threadify(insert_into_vfs, cache_result=True)(
                self.vfs, queue, len(self.paths)
            )

            prefixes = set(p[0] for p in self.paths if p[0])
            if prefixes:
                queue.put((QueueCommand.ENSURE_PREFIXES, prefixes))

            for virtual_path, path in self.paths:
                threadify(walk_path)(queue, virtual_path, path)

            t()

            delta = datetime.now() - notification_start_dt
            if self.notifier:
                notification = notification.copy(
                    body=f"Finished rescanning, it took {delta}"
                )
                self.notifier.notify(notification)

            log.log(100, f"A rescan finished after {delta}")

    def thomas_list(self, item, path, depth=0, modified_since=None):
        return self.list(path, depth, modified_since)



import logging
from contextlib import contextmanager

import requests
import requests.utils
from django.db import models
from jsonfield import JSONField
from picklefield.fields import PickledObjectField

from ...locktracker import LockTracker
from ...utils import hash_string
from .bencode import bdecode, bencode

logger = logging.getLogger(__name__)
websession_locktracker = LockTracker()


class WebSessionManager(models.Manager):
    @contextmanager
    def get_session(self, app):
        logger.debug(f"Getting requests session for {app}")
        with websession_locktracker.get_path(app):
            session = requests.Session()
            try:
                obj = self.model.objects.get(app=app)
                session.cookies = requests.utils.cookiejar_from_dict(
                    obj.requests_cookies
                )
            except self.model.DoesNotExist:
                obj = self.model(app=app)

            yield session
            obj.requests_cookies = requests.utils.dict_from_cookiejar(session.cookies)
            obj.save()


class WebSession(models.Model):
    app = models.CharField(max_length=30, unique=True)
    requests_cookies = PickledObjectField()
    last_modified = models.DateTimeField(auto_now=True)

    objects = WebSessionManager()


class TorrentFile(models.Model):
    app = models.CharField(max_length=30)
    url = models.CharField(max_length=5000)
    url_hash = models.CharField(max_length=50)
    torrent_data = models.BinaryField()
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (("app", "url_hash"),)

    def get_torrent_data(self):
        return bdecode(bytes(self.torrent_data))

    def get_infohash(self):
        return hash_string(bencode(self.get_torrent_data()[b"info"]))


class ListingCache(models.Model):
    app = models.CharField(max_length=30)
    search_token = models.CharField(max_length=50)
    path = models.CharField(max_length=500, default="")
    listing = JSONField()
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (("app", "search_token", "path"),)


class AvailableMedia(models.Model):
    app = models.CharField(max_length=30)
    identifier = models.CharField(max_length=500)
    created = models.DateTimeField(auto_now_add=True, db_index=True)

    class Meta:
        unique_together = (("app", "identifier"),)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.notifiers.multinotifier"
    verbose_name = "MultiNotifier Notifier"
    label = "notifier_multinotifier"

    def ready(self):
        from .handler import MultiNotifierNotifierHandlerPlugin  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.searchers.remotesearcher"
    verbose_name = "RemoteSearcher Searcher"
    label = "searchers_remotesearcher"

    def ready(self):
        from .handler import RemoteSearcherSearcher  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.config"
    verbose_name = "Config Service"
    label = "services_config"

    def ready(self):
        from .handler import ConfigServicePlugin  # NOQA
        from .configservice import ConfigConfigPlugin  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.available"
    verbose_name = "Available Metadata"
    label = "metadata_available"

    def ready(self):
        from .handler import AvailableMetadataHandlerPlugin  # NOQA



import json
import logging

from channels.exceptions import AcceptConnection, DenyConnection
from channels.generic.websocket import WebsocketConsumer
from channels.http import AsgiRequest
from django.contrib.auth import login
from jsonrpc import JSONRPCResponseManager
from jsonrpc.dispatcher import Dispatcher
from rest_framework import exceptions, permissions
from rest_framework.settings import api_settings
from unplugged import CascadingPermission

logger = logging.getLogger(__name__)


class DRFWebsocketConsumer(WebsocketConsumer):
    authentication_classes = api_settings.DEFAULT_AUTHENTICATION_CLASSES
    permission_classes = api_settings.DEFAULT_PERMISSION_CLASSES

    def get_authenticators(self):
        return [auth() for auth in self.authentication_classes]

    def get_permissions(self):
        return [permission() for permission in self.permission_classes]

    def check_permissions(self, request):
        for permission in self.get_permissions():
            if permission.has_permission(request, self):
                return True
        return False

    def authenticate(self, request):
        authenticators = self.get_authenticators()

        for authenticator in authenticators:
            try:
                result = authenticator.authenticate(request)
                if not result:
                    continue
                user, _ = result
                login(request, user)
                self.scope["user"] = user
                request.user = user
            except exceptions.AuthenticationFailed:
                continue
            else:
                break

    def connect(self):
        self.service = self.scope["service"]

        scope = dict(self.scope)
        scope["method"] = "get"
        request = AsgiRequest(scope, b"")
        request._request = request
        request.user = self.scope["user"]
        request.session = self.scope["session"]

        if not self.scope["user"].is_authenticated:
            self.authenticate(request)

        if self.check_permissions(request):
            raise AcceptConnection()
        else:
            raise DenyConnection()


class PlayerConsumer(DRFWebsocketConsumer):
    user = None
    permission_classes = (permissions.IsAuthenticated,)
    player_id = None
    controller = False

    def accept(self):
        super(PlayerConsumer, self).accept()
        self.user = self.scope["user"]
        self.player_coordinator = self.service.get_player_coordinator(self.user)

    def receive(self, text_data=None, bytes_data=None):
        logger.info(f"Got message from user {self.user!r}: {text_data}")
        response = JSONRPCResponseManager.handle(text_data, self.dispatcher).data
        return self.send(text_data=json.dumps(response))

    @property
    def dispatcher(self):  # TODO: make decorator or something
        d = Dispatcher()

        d.add_method(self.request_player_state, name="request_state")
        d.add_method(self.player_change_state, name="state")
        d.add_method(self.register_player, name="register")
        d.add_method(self.register_controller, name="subscribe")
        d.add_method(self.player_command, name="command")

        return d

    def register_player(self, player_id, name, commands, options):
        logger.debug(f"Adding a player:{player_id} for {self.user.username}")
        self.player_id = player_id
        self.player_coordinator.add_player(self, player_id, name, commands, options)

    def register_controller(self):
        logger.debug(f"Adding a controller for {self.user.username}")
        self.controller = True
        self.player_coordinator.add_controller(self)

    def player_command(self, player_id, method, *args):
        logger.debug(f"Calling command {method} on player_id {player_id}")
        player = self.player_coordinator.get_player(player_id)
        player.command(method, *args)

    def player_change_state(self, state, values, viewstate_id=None):
        player = self.player_coordinator.get_player(self.player_id)
        if not player:
            logger.info(f"Unknown player {self.player_id} found, dropping")
            self.close()
            return

        player.change_state(state, values, viewstate_id)

    def request_player_state(self, player_id, state, values):
        logger.debug("Called request player state")
        player = self.player_coordinator.get_player(player_id)
        player.request_state(state, values)

    def disconnect(self, message, **kwargs):
        logger.debug(f"User connection lost {self.user!r}")

        if self.controller:
            self.player_coordinator.remove_controller(self)

        if self.player_id:
            self.player_coordinator.remove_player(self.player_id)

    def jsonrpc_method(self, method, *args):
        self.send(
            text_data=json.dumps({"jsonrpc": "2.0", "method": method, "params": args})
        )



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.player"
    verbose_name = "Player Management Service"
    label = "services_player"

    def ready(self):
        from .handler import PlayerServicePlugin  # NOQA



import logging

from django.apps import AppConfig as DjangoAppConfig

logger = logging.getLogger(__name__)


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.mal"
    verbose_name = "MAL Metadata"
    label = "metadata_mal"

    def ready(self):
        from .parser import MALMetadataParserPlugin  # NOQA

        try:
            from .handler import MALMetadataHandlerPlugin  # NOQA
        except ImportError:
            logger.warning(
                "Unable to load mal metadata plugin, please install malparser"
            )



import logging
import re
from datetime import date

from django.db import models

from ...bases.metadata.models import (
    BaseListingItemRelation,
    BaseMetadata,
    BaseMetadataResolutionLink,
)

logger = logging.getLogger(__name__)

THETVDB_APIKEY = "E86EAF8A9978C847"


class Language(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Genre(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Network(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class ContentRating(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Status(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Metadata(BaseMetadata):
    imdb_id = models.IntegerField(null=True)

    rating = models.DecimalField(null=True, decimal_places=2, max_digits=4)
    votes = models.IntegerField(null=True)
    duration = models.IntegerField(null=True)
    year = models.IntegerField(null=True)
    synopsis = models.TextField(null=True)

    network = models.ForeignKey(Network, null=True, on_delete=models.deletion.CASCADE)
    content_rating = models.ForeignKey(
        ContentRating, null=True, on_delete=models.deletion.CASCADE
    )
    primary_language = models.ForeignKey(
        Language, null=True, on_delete=models.deletion.CASCADE
    )
    status = models.ForeignKey(Status, null=True, on_delete=models.deletion.CASCADE)

    genres = models.ManyToManyField(Genre)

    @property
    def metadata_name(self):
        return "thetvdb"

    def populate(self, config):
        import tvdb_api

        image_cache = config["image_cache"]

        tvdb = tvdb_api.Tvdb(apikey=THETVDB_APIKEY, banners=True)
        identifier = int(self.identifier)
        tv_show = tvdb[identifier]

        if tv_show.data["seriesName"] == "** 403: Series Not Permitted **":
            return

        self.genres.clear()

        self.network = None
        self.content_rating = None
        self.primary_language = None
        self.status = None
        self.imdb_id = None

        self.title = tv_show.data["seriesName"]
        self.rating = tv_show.data["siteRating"]
        self.votes = tv_show.data["siteRatingCount"]
        self.duration = tv_show.data["runtime"]

        if tv_show.data.get("imdb_id"):
            imdb_id = re.findall("tt(\d+)", tv_show.data["imdbId"])
            if imdb_id:
                self.imdb_id = imdb_id[0]

        self.year = None
        if tv_show.data.get("firstAired"):
            self.year = tv_show.data.get("firstAired").split("-")[0]

        self.synopsis = tv_show.data["overview"]

        if (
            tv_show.data.get("_banners", {}).get("poster", {}).get("raw")
        ):  # this is retarded design
            posters = sorted(
                tv_show.data["_banners"]["poster"]["raw"],
                key=lambda x: x["ratingsInfo"]["count"],
            )
            if posters:
                best_id = posters[-1]["id"]
                for _, v in tv_show.data["_banners"]["poster"].items():
                    if best_id in v:
                        self.cover = v[best_id]["_bannerpath"]
                        image_cache.get_image_path(self.cover)
                        break

        if tv_show.data.get("network"):
            self.network, _ = Network.objects.get_or_create(
                name=tv_show.data["network"]
            )

        if tv_show.data.get("rating"):
            self.content_rating, _ = ContentRating.objects.get_or_create(
                name=tv_show.data["rating"]
            )

        if tv_show.data.get("language"):
            self.primary_language, _ = Language.objects.get_or_create(
                name=tv_show.data["language"]
            )

        if tv_show.data.get("status"):
            self.status, _ = Status.objects.get_or_create(name=tv_show.data["status"])

        if tv_show.data["genre"]:
            for genre in tv_show.data["genre"]:
                genre, _ = Genre.objects.get_or_create(name=genre)
                self.genres.add(genre)

        Episode.objects.filter(metadata=self).delete()

        for season, episodes in tv_show.items():
            for episode, episode_info in episodes.items():
                e = Episode(metadata=self, season=season, episode=episode)
                e.title = episode_info["episodeName"]
                if episode_info.get("firstAired"):
                    e.air_date = date(
                        *[int(x) for x in episode_info["firstAired"].split("-")]
                    )

                e.save()


class Episode(models.Model):
    metadata = models.ForeignKey(Metadata, on_delete=models.deletion.CASCADE)
    season = models.IntegerField()
    episode = models.IntegerField()

    title = models.CharField(max_length=200, null=True)
    air_date = models.DateField(null=True)

    class Meta:
        unique_together = (("metadata", "season", "episode"),)


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(
        Metadata, db_index=True, on_delete=models.deletion.CASCADE
    )


THETVDB_LANGAUGE_REPLACEMENT_MAPPING = {
    "da": {"ae": "\xe6", "oe": "\xf8", "aa": "\xe5"}
}


class MetadataResolutionLink(BaseMetadataResolutionLink):
    metadata = models.ForeignKey(
        Metadata,
        db_index=True,
        null=True,
        blank=True,
        on_delete=models.deletion.CASCADE,
    )

    def resolve(self, config):
        logger.debug(f"Resolving using config {config!r}")

        language = config.get("search_language") or "en"

        search_strings = self.build_search_strings(
            THETVDB_LANGAUGE_REPLACEMENT_MAPPING.get(language, {})
        )
        if not search_strings:
            return

        import tvdb_api
        from tvdb_exceptions import tvdb_shownotfound

        tvdb = tvdb_api.Tvdb(apikey=THETVDB_APIKEY, language=language)
        for title, year in search_strings:
            try:
                tv_show = tvdb[title]
            except tvdb_shownotfound:
                continue

            return int(tv_show["id"])



import logging

from django.db import models

from ...bases.metadata.models import (
    BaseListingItemRelation,
    BaseMetadata,
    BaseMetadataResolutionLink,
)

logger = logging.getLogger(__name__)


class Language(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Country(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Genre(models.Model):
    name = models.CharField(max_length=50, unique=True)

    def __str__(self):
        return self.name


class Person(models.Model):
    identifier = models.CharField(max_length=100, unique=True)
    name = models.CharField(max_length=200)

    def __str__(self):
        return self.name


class Metadata(BaseMetadata):
    rating = models.DecimalField(null=True, decimal_places=2, max_digits=3)
    votes = models.IntegerField(null=True)
    duration = models.IntegerField(null=True)
    year = models.IntegerField(null=True)

    plot = models.TextField(null=True)
    synopsis = models.TextField(null=True)

    actors = models.ManyToManyField(Person, related_name="actors")
    writers = models.ManyToManyField(Person, related_name="writers")
    directors = models.ManyToManyField(Person, related_name="directors")

    genres = models.ManyToManyField(Genre)

    languages = models.ManyToManyField(Language)
    primary_language = models.ForeignKey(
        Language,
        null=True,
        related_name="primary_language",
        on_delete=models.deletion.CASCADE,
    )

    countries = models.ManyToManyField(Country)

    @property
    def metadata_name(self):
        return "imdb"

    def populate(self, config):
        from imdbparser import IMDb

        image_cache = config["image_cache"]

        imdb = IMDb()
        movie = imdb.get_movie(self.identifier)
        movie.fetch()

        self.actors.clear()
        self.writers.clear()
        self.directors.clear()
        self.genres.clear()
        self.languages.clear()
        self.countries.clear()

        AlternativeTitle.objects.filter(metadata=self).delete()

        self.title = movie.title
        self.rating = movie.rating
        self.votes = movie.votes
        self.duration = movie.duration
        self.year = movie.year
        self.plot = movie.plot
        self.synopsis = movie.description

        if movie.cover:
            self.cover = movie.cover
            image_cache.get_image_path(self.cover)

        person_target_map = [
            (movie.actors, self.actors),
            (movie.writers, self.writers),
            (movie.directors, self.directors),
        ]

        for persons, target in person_target_map:
            for p in persons:
                person, _ = Person.objects.get_or_create(
                    identifier=p.imdb_id, defaults={"name": p.name}
                )
                if p.name != person.name:
                    person.name = p.name
                    person.save()

                target.add(person)

        for genre in movie.genres:
            genre, _ = Genre.objects.get_or_create(name=genre)
            self.genres.add(genre)

        primary_language = None
        for language in movie.languages:
            language, _ = Language.objects.get_or_create(name=language)
            if not primary_language:
                primary_language = language
            self.languages.add(language)

        self.primary_language = primary_language

        for country in movie.countries:
            country, _ = Country.objects.get_or_create(name=country)
            self.countries.add(country)

        for alt_title in movie.alternative_titles:
            if alt_title:
                AlternativeTitle.objects.create(title=alt_title, metadata=self)


class AlternativeTitle(models.Model):
    title = models.CharField(max_length=500)
    metadata = models.ForeignKey(Metadata, on_delete=models.deletion.CASCADE)


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(
        Metadata, db_index=True, on_delete=models.deletion.CASCADE
    )


class MetadataResolutionLink(BaseMetadataResolutionLink):
    metadata = models.ForeignKey(
        Metadata,
        db_index=True,
        null=True,
        blank=True,
        on_delete=models.deletion.CASCADE,
    )

    def resolve(self, config):
        logger.debug(f"Resolving using config {config!r}")

        search_strings = self.build_search_strings({})
        if not search_strings:
            return

        from imdbparser import IMDb

        imdb = IMDb()

        for title, year in set(search_strings):
            logger.debug(f"Resolving using title:{title} year:{year}")

            if self.search_resolve == "tv":
                metadata = imdb.resolve_tv_show(title, year)
            elif self.search_resolve == "movie":
                metadata = imdb.resolve_movie(title, year)
            else:
                logger.warning(f"Unknown search resolve {self.search_resolve}")
                continue

            if not metadata:
                continue

            return metadata.imdb_id



import os
import shutil
import tempfile
import unittest

from django.test import TestCase

from ...plugins import DatabaseCacheLayer
from .handler import ShelfDatabasePlugin


class ShelfDatabaseTest(TestCase):
    def setUp(self):
        self.temp_path = tempfile.mkdtemp()
        self.db_path = os.path.join(self.temp_path, "db")
        try:
            self.db = self.open_database()
        except ImportError:
            raise unittest.SkipTest("Database not installed")

    def tearDown(self):
        shutil.rmtree(self.temp_path)

    def open_database(self):
        return ShelfDatabasePlugin({"path": self.db_path})

    def reOpen(self):
        self.db.sync()
        self.db.close()
        self.db = self.open_database()

    def test_set_get_keys(self):
        self.db["\x01\x03"] = "\x01\x03"
        self.db["test"] = "test"

        self.db.sync()

        self.assertEqual(self.db["\x01\x03"], "\x01\x03")
        self.assertEqual(self.db["test"], "test")
        self.assertEqual(len(self.db), 2)

    def test_delete_keys(self):
        self.test_set_get_keys()

        del self.db["test"]
        self.db.sync()

        self.assertEqual(len(self.db), 1)
        self.assertNotIn("test", self.db)
        self.assertIn("\x01\x03", self.db)

    def test_reopen_database(self):
        self.db["\x01\x03"] = "\x01\x03"
        self.db["test"] = "test"

        self.reOpen()

        self.assertEqual(self.db["test"], "test")
        self.assertEqual(len(self.db), 2)
        self.assertEqual(self.db["\x01\x03"], "\x01\x03")
        self.db["test2"] = "abc"

        del self.db["test"]

        self.reOpen()

        self.assertEqual(len(self.db), 2)
        self.assertNotIn("test", self.db)
        self.assertIn("\x01\x03", self.db)
        self.assertIn("test2", self.db)

    def test_complex_types(self):
        self.db["set"] = set([1, 2, 3])
        self.assertEqual(self.db["set"], set([1, 2, 3]))

        self.db["dict"] = dict(a=1, b=2)
        self.assertEqual(self.db["dict"], dict(a=1, b=2))


class ShelfDatabaseCacheTest(ShelfDatabaseTest):
    def open_database(self):
        return DatabaseCacheLayer(super(ShelfDatabaseCacheTest, self).open_database())



import logging

from guessit import guessit
from unplugged import Schema

from .plugin import ItemInfoPlugin

logging.getLogger("rebulk.rebulk").setLevel(logging.WARNING)


class GuessItItemInfoPlugin(ItemInfoPlugin):
    plugin_name = "guessit"
    config_schema = Schema

    def get_info(self, name):
        info = dict(guessit(name))

        retval = {}
        for k, v in info.items():
            if k in ["season", "episode"]:
                if isinstance(v, list):
                    v = v[-1]
                retval[k] = v
            elif k == "episode_details":
                retval["episode_type"] = v
            elif k == "format":
                retval["source"] = v
            elif k == "part":
                if isinstance(v, list):
                    v = v[-1]
                retval["episode"] = v
            elif k in ["title", "year"]:
                retval[k] = v

        lower_name = name.lower()
        types = {
            "nfofix": ["nfofix", "nfo.fix", "nfo_fix"],
            "dirfix": ["dirfix", "dir.fix", "dir_fix"],
            "subfix": ["subfix", "sub.fix", "sub_fix"],
            "subpack": ["subpack", "sub.pack", "sub_pack"],
        }
        for k, vs in types.items():
            for v in vs:
                if v in lower_name:
                    retval["fix_type"] = k

        return retval

    def unload(self):
        pass



import logging

from django.http import Http404
from rest_framework import status
from rest_framework.response import Response
from unplugged import CascadingPermission, JSONAPIObject, JSONAPIRoot

from ...bases.listing.views import BaseListingView

logger = logging.getLogger(__name__)


class SectionsListView(BaseListingView):
    permission_classes = (CascadingPermission,)

    def get_config_view(self, request):
        root = JSONAPIRoot()
        for section in self.service.config["sections"]:
            logger.debug(
                f"Making config view {section['name']} with settings {section!r}"
            )

            links = {
                "self": request.build_absolute_uri(
                    f"/{self.service.name}/{section['name']}"
                )
            }
            obj = JSONAPIObject(
                "folder", f"{self.service.name}/{section['name']}", links=links
            )
            obj["name"] = section["name"]
            if section.get("display_name"):
                obj["display_name"] = section["display_name"]

            first_level_config = sorted(section["levels"], key=lambda x: x["depth"])[0]
            filter_info_obj = self.get_level_filter_info(
                obj["name"], first_level_config
            )
            obj.add_relationship("metadata_filterinfo", filter_info_obj, local=True)

            root.append(obj)

        return root.serialize(request)

    def add_additional_parent_relationships(
        self, parent, request, config, listing_item_root
    ):
        parent_filter_info_obj = self.get_level_filter_info(
            listing_item_root.path, config["level"]
        )
        parent.add_relationship("metadata_filterinfo", parent_filter_info_obj)



import logging
from collections import defaultdict

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from unplugged import Schema

from ...plugins import MetadataHandlerPlugin
from .mixins import (
    GetMetadataStatsMixin,
    ListingItemRelinkingMixin,
    PopulateMetadataJSONAPIMixin,
    ResetFailedMixin,
)

logger = logging.getLogger(__name__)


class LinkingMetadataHandlerPlugin(
    ListingItemRelinkingMixin,
    PopulateMetadataJSONAPIMixin,
    GetMetadataStatsMixin,
    ResetFailedMixin,
    MetadataHandlerPlugin,
):
    linkable = False

    metadata_link_model = None
    user_field = None
    config_schema = Schema

    def link_metadata_listingitems(self, listing_build, fetch_metadata=True):
        listing_item_root = listing_build.listing_item_root
        linkable_metadata = listing_build.linkable_metadata

        metadata_mapping = defaultdict(set)
        for p in linkable_metadata:
            base_qs = p.listing_item_relation_model.objects.filter(
                Q(listingitem__parent=listing_item_root)
                | Q(listingitem=listing_item_root)
            )
            items_to_link = self.metadata_link_model.objects.filter(
                content_type=ContentType.objects.get_for_model(p.model),
                object_id__in=base_qs.values("metadata_id"),
            )

            local_metadata_mapping = defaultdict(list)
            if self.user_field:
                for pk, object_id, user_id in items_to_link.values_list(
                    "pk", "object_id", self.user_field
                ):
                    local_metadata_mapping[object_id].append((pk, user_id))
            else:
                for pk, object_id in items_to_link.values_list("pk", "object_id"):
                    local_metadata_mapping[object_id].append((pk, None))

            for listingitem_id, metadata_id in base_qs.filter(
                metadata_id__in=items_to_link.values("object_id")
            ).values_list("listingitem_id", "metadata_id"):
                for item in local_metadata_mapping[metadata_id]:
                    metadata_mapping[listingitem_id].add(item)

        self.listingitem_relinking(listing_item_root, metadata_mapping)

    def schedule_update(self):
        pass

    def unload(self):
        pass



# Generated by Django 2.2.8 on 2019-12-13 16:29

import django.db.models.deletion
import jsonfield.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="DefaultSetting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "default_type",
                    models.CharField(
                        choices=[
                            ("system", "System Default"),
                            ("configured", "Configured Default"),
                        ],
                        max_length=30,
                    ),
                ),
                ("namespace", models.CharField(max_length=100)),
                ("key", models.CharField(max_length=100)),
                ("value", jsonfield.fields.JSONField()),
            ],
            options={"unique_together": {("default_type", "namespace", "key")},},
        ),
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("namespace", models.CharField(max_length=100)),
                ("key", models.CharField(max_length=100)),
                ("value", jsonfield.fields.JSONField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"unique_together": {("user", "namespace", "key")},},
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-13 16:29

import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SearchQueryCache",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("query_hash", models.CharField(max_length=50, unique=True)),
                ("query", jsonfield.fields.JSONField()),
                ("last_used", models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



import logging

from django.db import models

from ...bases.metadata.models import BaseListingItemRelation, BaseMetadata

logger = logging.getLogger(__name__)


class Metadata(BaseMetadata):
    name = models.CharField(max_length=255)

    @property
    def metadata_name(self):
        return "name"

    def populate(self, config):
        Metadata.objects.filter(last_update_status="pending").update(
            last_update_status="success"
        )


class ListingItemRelation(BaseListingItemRelation):
    metadata = models.ForeignKey(
        Metadata, db_index=True, on_delete=models.deletion.CASCADE
    )



from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from jsonfield import JSONField
from unplugged.models import Plugin

from ..listing.models import ListingItem

COVERS_FOLDER = "covers"


class BaseUpdatable(models.Model):
    priority = models.SmallIntegerField(default=20)

    UPDATE_STATUS_DO_NOT_FETCH = (
        "do-not-fetch"  # be able to link, but not actually needed right now
    )
    UPDATE_STATUS_PENDING = "pending"
    UPDATE_STATUS_UPDATING = "updating"
    UPDATE_STATUS_FAILED = "failed"
    UPDATE_STATUS_SUCCESS = "success"

    UPDATE_STATUS_CHOICES = (
        (UPDATE_STATUS_DO_NOT_FETCH, "Do not fetch"),
        (UPDATE_STATUS_PENDING, "Pending"),
        (UPDATE_STATUS_UPDATING, "Updating"),
        (UPDATE_STATUS_FAILED, "Failed"),
        (UPDATE_STATUS_SUCCESS, "Success"),
    )
    last_update_status = models.CharField(
        max_length=20, choices=UPDATE_STATUS_CHOICES, default="pending", db_index=True
    )

    last_updated = models.DateTimeField(null=True)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True


class BaseMetadata(BaseUpdatable):
    """Used by remote metadata downloaders as base"""

    identifier = models.CharField(max_length=255, unique=True)
    populated = models.BooleanField(default=False)

    plugin = models.ForeignKey(
        Plugin,
        null=True,
        blank=True,
        related_name="metadata_%(app_label)s",
        on_delete=models.deletion.CASCADE,
    )

    title = models.CharField(max_length=300, null=True)
    cover = models.URLField(null=True, max_length=1000)

    prefetch_related_denormalized = JSONField(blank=True, null=True)

    @property
    def metadata_name(self):
        raise NotImplementedError

    def get_absolute_url(self):
        if not hasattr(self, "__config__"):
            return

        if not self.__config__.get("metadata_server"):
            return

        metadata_server = self.__config__["metadata_server"]
        return reverse(
            f"unplugged:{metadata_server.name}:metadata",
            kwargs={
                "metadata_handler": self.metadata_name,
                "identifier": self.identifier,
            },
        )

    def populate(self, config):
        raise NotImplementedError

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.identifier} in {self.__class__.__name__}"


class BaseListingItemRelation(models.Model):
    listingitem = models.ForeignKey(
        ListingItem,
        related_name="%(app_label)s",
        db_index=True,
        on_delete=models.deletion.CASCADE,
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="%(app_label)s",
        null=True,
        on_delete=models.deletion.CASCADE,
    )

    class Meta:
        abstract = True
        unique_together = (("listingitem", "metadata", "user"),)


class BaseMetadataLink(models.Model):
    """Used to link with metadata and add an additional field"""

    content_type = models.ForeignKey(ContentType, on_delete=models.deletion.CASCADE)
    object_id = models.PositiveIntegerField()
    metadata = GenericForeignKey()

    class Meta:
        abstract = True
        unique_together = (("content_type", "object_id"),)


class BaseMetadataResolutionLink(BaseUpdatable):
    """Used for auto-resolving a name to metadata when identifier is missing"""

    name = models.CharField(max_length=255)
    search_resolve = models.CharField(max_length=30)

    listingitems = models.ManyToManyField(
        ListingItem, related_name="resolution_%(app_label)s"
    )
    plugin = models.ForeignKey(
        Plugin,
        null=True,
        blank=True,
        related_name="resolution_%(app_label)s",
        on_delete=models.deletion.CASCADE,
    )

    class Meta:
        abstract = True
        unique_together = (("name", "search_resolve"),)

    def __str__(self):
        return "%s in %s" % (self.name, self.__class__.__name__)

    def build_search_strings(self, permutations):
        search_strings = []
        for listingitem in self.listingitems.all():  # TODO: something is wrong here
            iteminfos = listingitem.get_metadata("iteminfo") or []
            for iteminfo in iteminfos:
                title = iteminfo.title
                if title:
                    search_strings.append((title, iteminfo.year))

            if not search_strings:
                name_search_string = listingitem.name.strip()
                if " " not in name_search_string:
                    name_search_string = name_search_string.replace(".", " ").replace(
                        "_", " "
                    )

                search_strings.append((name_search_string, None))

            first_search_string = search_strings[0]
            for src, dst in permutations.items():
                search_strings.append(
                    (first_search_string[0].replace(src, dst), first_search_string[1])
                )

        seen = set()
        return [x for x in search_strings if not (x in seen or seen.add(x))]

    def populate(self, config):
        raise NotImplementedError



from ..shelve.tests import ShelfDatabaseTest
from .handler import LevelDBDatabasePlugin


class LevelDBDatabaseTest(ShelfDatabaseTest):
    def open_database(self):
        return LevelDBDatabasePlugin({"path": self.db_path})



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.inputs.rfs"
    verbose_name = "RemoteFilesystem Input"
    label = "inputs_rfs"

    def ready(self):
        from .handler import RemoteFilesystemInputPlugin  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.bases.websearcher"
    verbose_name = "WebSearcher base"
    label = "searcher_websearcher_base"



# Generated by Django 2.2.8 on 2019-12-13 16:29

import django.db.models.deletion
import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ListingItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app", models.CharField(max_length=30)),
                ("name", models.CharField(db_index=True, default="", max_length=200)),
                ("path", models.CharField(max_length=500)),
                ("datetime", models.DateTimeField()),
                ("config", jsonfield.fields.JSONField(default=dict)),
                ("attributes", jsonfield.fields.JSONField(default=dict)),
                (
                    "item_type",
                    models.CharField(
                        choices=[("file", "File"), ("folder", "Folder")], max_length=10
                    ),
                ),
                ("is_root", models.BooleanField(default=False)),
                ("last_updated", models.DateTimeField(null=True)),
                ("last_checked", models.DateTimeField(null=True)),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="services_listing.ListingItem",
                    ),
                ),
            ],
            options={"ordering": ("datetime",), "unique_together": {("app", "path")},},
        ),
    ]



import logging

from django.http import Http404
from rest_framework import permissions
from rest_framework.response import Response
from rest_framework.views import APIView
from unplugged import CascadingPermission, JSONAPIObject, JSONAPIRoot

logger = logging.getLogger(__name__)


class MetadataView(APIView):
    service = None
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, metadata_handler, identifier):
        logger.debug(
            f"Fetching metadata using {metadata_handler} with identifier {identifier}"
        )
        for plugin in self.service._related_plugins:
            if (
                plugin.name == metadata_handler
                and plugin.plugin_type == "metadatahandler"
            ):
                break
        else:
            logger.warning(f"Did not find metadata_handler {metadata_handler}")
            raise Http404

        metadata = plugin.get_metadata(request, identifier)
        if not metadata:
            raise Http404

        root = JSONAPIRoot()

        identifier = f"{plugin.plugin_name}:{identifier}"
        obj_type = plugin.get_jsonapi_type()
        obj = JSONAPIObject(obj_type, identifier)
        obj.update(metadata)
        root.append(obj)

        return Response(root.serialize(request))



import pickle

from unplugged import Schema, fields

import lmdb

from ...plugins import DatabasePlugin


class LMDBDatabaseSchema(Schema):
    path = fields.String()


class LMDBDatabasePlugin(DatabasePlugin):
    plugin_name = "lmdb"

    config_schema = LMDBDatabaseSchema

    def __init__(self, config):
        self.db = lmdb.open(
            self.get_database_path(config["path"]),
            map_size=500 * 1024 * 1024,  # TODO: autoscale
            map_async=True,
            writemap=True,
            metasync=False,
        )

    def unload(self):
        self.close()

    def __getitem__(self, key):
        with self.db.begin() as txn:
            value = txn.get(key.encode("utf-8"))
            if value:
                return pickle.loads(value)

    def __setitem__(self, key, value):
        with self.db.begin(write=True) as txn:
            txn.put(key.encode("utf-8"), pickle.dumps(value))

    def __delitem__(self, key):
        with self.db.begin(write=True) as txn:
            txn.delete(key.encode("utf-8"))

    def keys(self):
        return [k for k, v in self.db.cursor()]

    def __len__(self):
        return int(self.db.stat()["entries"])

    def __contains__(self, key):
        return self[key] is not None

    def close(self):
        self.db.close()

    def sync(self):
        self.db.sync()

    def __iter__(self):
        return self.db.cursor()



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.history"
    verbose_name = "History Metadata"
    label = "metadata_history"

    def ready(self):
        from .handler import HistoryMetadataHandlerPlugin  # NOQA
        from .history import HistoryHistoryPlugin  # NOQA



import logging
from threading import Lock

import tvdb_api  # NOQA
from rest_framework import serializers
from unplugged import fields

from ...bases.metadata.remotemetadata import (
    MetadataSerializer,
    RemoteMetadataHandlerPlugin,
)
from ...bases.metadata.schema import MetadataSchema
from .filters import MetadataFilter
from .models import ListingItemRelation, Metadata, MetadataResolutionLink

logger = logging.getLogger(__name__)


class TheTVDBMetadataSchema(MetadataSchema):
    search_resolve = fields.Boolean()
    search_language = fields.String(
        enum=["en", "da"], enum_names=["English", "Danish"]
    )  # choices, should also try to change language to local names
    episodic = fields.Boolean()  # Resolve to episodes


class TheTVDBMetadataSerializer(MetadataSerializer):
    primary_language = serializers.StringRelatedField()
    genres = serializers.StringRelatedField(many=True)

    class Meta(MetadataSerializer.Meta):
        model = Metadata


class TheTVDBMetadataHandlerPlugin(RemoteMetadataHandlerPlugin):
    plugin_name = "thetvdb"
    select_related = [
        "primary_language",
        "network",
        "content_rating",
        "primary_language",
        "status",
    ]
    prefetch_related = ["genres"]
    currently_updating_lock = Lock()
    serializer = TheTVDBMetadataSerializer
    model = Metadata
    filter = MetadataFilter
    listing_item_relation_model = ListingItemRelation
    metadata_resolution_link_model = MetadataResolutionLink
    fulltext_fields = ["title"]

    config_schema = TheTVDBMetadataSchema

    simpleadmin_templates = [
        {
            "template": {
                "search_resolve": True,
                "search_language": {"simpleAdminMethod": "userInput"},
            },
            "description": "Fetch metadata from TheTVDB",
            "id": "default",
            "update_method": "full",
            "display_name": "TheTVDB",
        }
    ]

    __traits__ = ["primary_metadata", "metadata_tv"]



import logging
import os
from shelve import DbfilenameShelf

from unplugged import Schema, fields

from ...plugins import DatabasePlugin

logger = logging.getLogger(__name__)


class ShelfDatabaseSchema(Schema):
    path = fields.String()


class ShelfDatabasePlugin(DbfilenameShelf, DatabasePlugin):
    plugin_name = "shelf"

    config_schema = ShelfDatabaseSchema

    _doing_sync = False

    def __init__(self, config):
        self.config = config
        self._open()

    def _open(self):
        DbfilenameShelf.__init__(
            self,
            os.path.join(self.get_database_path(self.config["path"]), "shelfdb.db"),
        )

    def unload(self):
        self.close()

    def __repr__(self):
        return DatabasePlugin.__repr__(self)

    def sync(self):
        if self._doing_sync:
            return

        if self.writeback and self.cache:
            self.writeback = False
            for key, entry in self.cache.items():
                self[key] = entry
            self.writeback = True
            self.cache = {}

        if hasattr(self.dict, "reorganize"):
            # The database might grow too much without an occational reorganize
            # This is not the perfect place to put it but shelve is a bad choice to begin with
            self.dict.reorganize()

        if hasattr(self.dict, "sync"):  # "thanks" python
            self.dict.sync()
        else:
            logger.debug("Doing close/open database sync")
            self._doing_sync = True
            self.close()
            self._open()
            self._doing_sync = False
            logger.debug("Done with close/open database sync")



# Generated by Django 2.2.8 on 2020-03-31 08:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('metadata_embedded', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='embeddedinfo',
            name='group',
            field=models.CharField(max_length=300, null=True),
        ),
        migrations.AddField(
            model_name='embeddedinfo',
            name='mediainfo_audio',
            field=models.CharField(max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='embeddedinfo',
            name='mediainfo_dual_audio',
            field=models.BooleanField(default=False),
        ),
    ]



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.webinterface"
    verbose_name = "Webinterface Service"
    label = "services_webinterface"

    def ready(self):
        from .handler import WebinterfaceServicePlugin  # NOQA



from django.db import models
from jsonfield import JSONField

from ...plugins import SearchQuery


class SearchQueryCache(models.Model):
    query_hash = models.CharField(max_length=50, unique=True)
    query = JSONField()

    last_used = models.DateTimeField(auto_now=True)

    def get_search_query(self, filters):
        return SearchQuery(filters, self.query)



import logging
from datetime import timedelta

from django.conf.urls import url
from django.utils.timezone import now
from unplugged import DefaultPermission, RelatedPluginField, Schema, fields

from ...bases.listing.handler import BaseListingService
from ...bases.listing.schema import BaseSchema, ListingSchema
from ...exceptions import NotModifiedException
from ...plugins import SearcherPlugin, SearcherPluginManager
from .models import SearchQueryCache
from .views import StoreListView

logger = logging.getLogger(__name__)
LISTING_CACHE_TIME = timedelta(minutes=10)


class StoreSearcherSchema(Schema):
    preference = fields.Integer(default=10)
    searcher = RelatedPluginField(plugin_type=SearcherPlugin)


class StoreSchema(ListingSchema):
    searchers = fields.Nested(StoreSearcherSchema, many=True, default=list)
    rebuild_automatically = fields.Boolean(default=False)


class StoreServiceSchema(BaseSchema):
    sections = fields.Nested(StoreSchema, many=True, default=list)


class StoreService(BaseListingService):
    plugin_name = "store"
    config_schema = StoreServiceSchema
    default_permission = DefaultPermission.ALLOW

    simpleadmin_templates = [
        {
            "template": {
                "display_name": {"simpleAdminMethod": "userInput", "required": True},
                "searchers": {"simpleAdminMethod": "userInput"},
                "name": {"simpleAdminMethod": "slugify", "source": "display_name"},
                "histories": [
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "history",
                        "name": "history",
                    }
                ],
                "rebuild_automatically": False,
                "levels": [
                    {
                        "indexer": None,
                        "depth": 1,
                        "content_type": "movies",
                        "listing_depth": 0,
                        "default_ordering": "metadata_embedded__index",
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "embedded",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "imdb",
                            },
                        ],
                    },
                    {
                        "indexer": None,
                        "depth": 2,
                        "content_type": "movie",
                        "listing_depth": 0,
                        "default_ordering": "metadata_embedded__index",
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "iteminfo",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "embedded",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "available",
                            },
                        ],
                    },
                ],
            },
            "description": "Search for movies",
            "id": "movies",
            "update_method": "modify_key",
            "modify_key": "sections",
        },
        {
            "template": {
                "display_name": {"simpleAdminMethod": "userInput", "required": True},
                "searchers": {"simpleAdminMethod": "userInput"},
                "name": {"simpleAdminMethod": "slugify", "source": "display_name"},
                "histories": [
                    {
                        "simpleAdminMethod": "lookupPlugin",
                        "plugin_type": "history",
                        "name": "history",
                    }
                ],
                "rebuild_automatically": False,
                "levels": [
                    {
                        "indexer": None,
                        "depth": 1,
                        "content_type": "tvshows",
                        "listing_depth": 0,
                        "default_ordering": "metadata_embedded__index",
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "embedded",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "imdb",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "mal",
                            },
                        ],
                    },
                    {
                        "indexer": None,
                        "depth": 2,
                        "content_type": "seasons",
                        "default_ordering": "-metadata_embedded__season,metadata_embedded__index",
                        "listing_depth": 0,
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "embedded",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "available",
                            },
                        ],
                    },
                    {
                        "indexer": None,
                        "depth": 3,
                        "content_type": "episodes",
                        "default_ordering": "-metadata_embedded__season,-metadata_embedded__episode,-metadata_iteminfo__season,-metadata_iteminfo__episode,metadata_embedded__index",
                        "listing_depth": 0,
                        "metadata_handlers": [
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "name",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "historymetadata",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "iteminfo",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "embedded",
                            },
                            {
                                "simpleAdminMethod": "lookupPlugin",
                                "plugin_type": "metadatahandler",
                                "name": "available",
                            },
                        ],
                    },
                ],
            },
            "description": "Search for TV Shows\n"
            "Data must be stored in show/season.01/episode.01 format",
            "id": "tv-season-episode",
            "update_method": "modify_key",
            "modify_key": "sections",
        },
    ]

    def get_urls(self):
        return [
            url(
                r"^(?P<path>.*)$",
                StoreListView.as_view(
                    service=self, listing_builder=self.listing_builder
                ),
            )
        ]

    def get_filters(self, config):
        return SearcherPluginManager.filters_multiple(
            [searcher["searcher"] for searcher in config["searchers"]]
        )

    def get_path_config(self, path):
        config = super(StoreService, self).get_path_config(path)
        config["fetch_metadata"] = False
        return config

    def get_item(self, config, path):
        logger.debug("Getting item for %s" % (path,))
        if not config.get("searchers"):
            return None
        path = path.split("/")
        root_path, query_hash = path[:2]
        trailing_path = "/".join(path[2:])
        try:
            search_query_cache = SearchQueryCache.objects.get(query_hash=query_hash)
        except SearchQueryCache.DoesNotExist:
            logger.info(f"Trying to list a non-existant query_hash {query_hash}")
            return None

        searchers = [searcher["searcher"] for searcher in config["searchers"]]

        if (
            trailing_path
        ):  # We cant create an item for a subpath, we need to have the original query
            return None

        filters = self.get_filters(config)
        search_query = search_query_cache.get_search_query(filters)
        item = SearcherPluginManager.get_item_multiple(
            searchers, query_hash, search_query
        )

        logger.debug(f"we are at {trailing_path} - {config!r}")

        if not trailing_path and "display_name" in config:
            item.id = config["display_name"]

        return item



import os

from django.conf import settings
from unplugged import JSONAPIObject


class Stream:
    def __init__(self, url, playhandler, name=None):
        self.url = url
        self.playhandler = playhandler
        self.name = name

    def to_jsonapi(self, request):
        obj = JSONAPIObject(
            f"stream_{self.playhandler}",
            self.name,
            links={"stream": request.build_absolute_uri(self.url)},
        )
        return obj

    def serialize(self, request):
        retval = {
            "url": request.build_absolute_uri(self.url),
            "playhandler": self.playhandler,
        }
        if self.name:
            retval["name"] = self.name

        return retval

    @classmethod
    def unserialize(cls, data):
        return cls(data["url"], data["playhandler"], data.get("name"))


def create_stream(item, request):
    stream_or_item = item.stream()
    if isinstance(stream_or_item, Stream):
        return stream_or_item
    else:
        url = request.build_absolute_uri(
            settings.THOMAS_HTTP_OUTPUT.serve_item(stream_or_item)
        )
        return Stream(url, "http", stream_or_item.id)



import rest_framework_filters as filters

from .models import ListingItemRelation


class MetadataFilter(filters.FilterSet):
    id = filters.AllLookupsFilter(
        field_name="metadata__identifier", name="metadata__identifier"
    )
    genre = filters.AllLookupsFilter(
        field_name="metadata__genres__name", name="metadata__genres__name"
    )  # choices
    rating = filters.AllLookupsFilter(
        field_name="metadata__rating", name="metadata__rating"
    )
    votes = filters.AllLookupsFilter(
        field_name="metadata__votes", name="metadata__votes"
    )
    duration = filters.AllLookupsFilter(
        field_name="metadata__duration", name="metadata__duration"
    )
    year = filters.AllLookupsFilter(field_name="metadata__year", name="metadata__year")
    primary_language = filters.AllLookupsFilter(
        field_name="metadata__primary_language__name",
        name="metadata__primary_language__name",
    )  # choices
    content_rating = filters.AllLookupsFilter(
        field_name="metadata__content_rating__name",
        name="metadata__content_rating__name",
    )  # choices
    status = filters.AllLookupsFilter(
        field_name="metadata__status__name", name="metadata__status__name"
    )  # choices
    network = filters.AllLookupsFilter(
        field_name="metadata__network__name", name="metadata__network__name"
    )  # choices

    class Meta:
        model = ListingItemRelation
        order_by = ["rating", "votes", "duration", "year"]
        fields = []
        include_related = [
            "primary_language",
            "genre",
            "network",
            "content_rating",
            "status",
        ]



import os

import django.contrib.admin
from django.conf import settings
from django.conf.urls import include, static, url
from django.contrib import admin
from django.urls import path, re_path

from tridentstream.utils import serve

urlpatterns = [
    path("djangoadmin/", admin.site.urls),
    url(r"^api-auth/", include("rest_framework.urls", namespace="rest_framework")),
    url(r"", include(("unplugged.urls", "unplugged"), namespace="unplugged")),
]



import time
from collections import defaultdict

from thomas import Item, router

from .exceptions import PathNotFoundException
from .plugins import DatabaseCacheLayer
from .utils import hash_string

COMMIT_COUNTER = 10000  # listing entries before adding to work queue
BUCKET_SIZE = 1  # bytes,
GC_DELETED_FILES = 60 * 60 * 24 * 30  # seconds


class AlreadyInSessionException(Exception):
    """There's a session already in progress, only one session at a time"""


class NotInSessionException(Exception):
    """Must always be in session when modifying the database"""


class DatabaseType:
    FILE = "\x00"
    DIRECTORY = "\x01"
    BUCKET = "\x02"
    FILELIST = "\x03"
    NEW_BUCKET = "\x04"
    DELETED_BUCKET = "\x05"

    _mapping = {"\x00": "file", "\x01": "folder"}

    @staticmethod
    def to_str(database_type):
        return DatabaseType._mapping[database_type]


def cleanup_path(path):
    if type(path) is list:
        return [cleanup_path(p) for p in path]
    else:
        return path.strip("/")


def keyify(key_type, path):
    return "%s%s" % (key_type, hash_string(cleanup_path(path)))


def merge_set_dicts(result, merge_with):
    for k, v in merge_with.items():
        result[k] |= v


def generate_buckets(bucket_depth=BUCKET_SIZE):
    for i in [chr(x) for x in range(256)]:
        if bucket_depth > 1:
            for j in generate_buckets(bucket_depth - 1):
                yield i + j
        else:
            yield i


def compare_dicts(src, dst):
    """Checks if all keys in src matches the values in dst"""
    for k, v in src.items():
        if k not in dst or v != dst[k]:
            return False

    return True


class FileSystem:
    _in_session = False
    _delete_missing_items = False
    _always_trigger_new = False
    _current_time = None

    def __init__(self, db):
        self.db = DatabaseCacheLayer(db)
        self.events = {
            "new": [],
            "deleted": [],
            "db_closed": [],
            "file_route_needed": [],
            "list_route_needed": [],
            "on_item": [],
        }

        self.initialize_database()
        self.reset_session()

    def add_event(self, event, f):
        self.events[event].append(f)

    def _trigger_event(self, event_name, *args, **kwargs):
        for event in self.events.get(event_name, []):
            event(*args, **kwargs)

    def close(self):
        self._trigger_event("db_closed")

    def initialize_database(self):
        for bucket in generate_buckets():
            key = "%s%s" % (DatabaseType.BUCKET, bucket)
            if key not in self.db:
                self.db[key] = set()

            key = "%s%s" % (DatabaseType.DELETED_BUCKET, bucket)
            if key not in self.db:
                self.db[key] = set()

            key = "%s%s" % (DatabaseType.NEW_BUCKET, bucket)
            self.db[key] = set()

        key = keyify(DatabaseType.DIRECTORY, "")
        if key not in self.db:
            self.db[key] = {"path": "", "date": 0}

            filelist_key = keyify(DatabaseType.FILELIST, "")
            self.db[filelist_key] = set()

    def reset_session(self):
        """
        Resets buckets to allow deleting files missing from the new session.

        This can be used to update a listing with new stuff and remove things not there anymore.
        """

        self.reset_buckets()
        self.reset_keys()

    def commit_session(self):
        """
        After a scan is done, this commits the new bucket session as the current bucket session.
        """

        self.commit_buckets()
        self.commit_keys()
        self.db.sync()

    def session(self, complete=False, always_trigger_new=False, current_time=None):
        self._delete_missing_items = complete
        self._always_trigger_new = always_trigger_new
        self._current_time = current_time
        return self

    def __enter__(self):
        """This should only be used"""
        if self._in_session:
            raise AlreadyInSessionException()

        self._current_time = (
            isinstance(self._current_time, int)
            and self._current_time
            or int(time.time())
        )
        self._in_session = True
        self.reset_session()

    def __exit__(self, type, value, traceback):
        self.commit_session()
        self.finish_buckets()
        self._delete_missing_items = False
        self._always_trigger_new = False
        self._current_time = None
        self._in_session = False

    def finish_buckets(self):
        deleted_paths = []
        for bucket in generate_buckets():
            new_key = "%s%s" % (DatabaseType.NEW_BUCKET, bucket)
            cur_key = "%s%s" % (DatabaseType.BUCKET, bucket)
            zombie_key = "%s%s" % (DatabaseType.DELETED_BUCKET, bucket)
            new_hashes = self.db[new_key]
            cur_hashes = self.db[cur_key] if not self._always_trigger_new else set()
            zombie_hashes = self.db[zombie_key]

            for key in new_hashes - cur_hashes:
                m = self.db[key]
                self._trigger_event(
                    "new", key=key, db_type=DatabaseType.to_str(key[0]), metadata=m
                )
                self.db[key] = m

            if self._delete_missing_items:
                deleted_hashes = cur_hashes - new_hashes

                for key in deleted_hashes:
                    m = self.db[key]
                    m["deleted"] = self._current_time
                    self._trigger_event(
                        "deleted",
                        key=key,
                        db_type=DatabaseType.to_str(key[0]),
                        metadata=m,
                    )
                    self.db[key] = m
                    deleted_paths.append(m["path"])

                self.db[cur_key] = new_hashes

                zombie_remove_keys = set()
                for key in zombie_hashes:
                    m = self.db[key]
                    if (
                        "deleted" in m
                        and m["deleted"] < self._current_time - GC_DELETED_FILES
                    ):
                        self._remove_key(key, m)
                        zombie_remove_keys.add(key)
                zombie_hashes -= zombie_remove_keys

                self.db[zombie_key] = (cur_hashes - new_hashes) | (
                    zombie_hashes - new_hashes
                )
            else:
                self.db[cur_key] = new_hashes | cur_hashes

            self.db[new_key] = set()

        last_touched_path = None
        for path in sorted(deleted_paths):
            if last_touched_path is not None and path.startswith(last_touched_path):
                continue

            self.add_parent_path_touched(path, self._current_time)
            last_touched_path = path

        if last_touched_path is not None:
            self.commit_keys()

        self.db.sync()

    def _remove_key(self, key, metadata):
        del self.db[key]
        if key[0] == DatabaseType.DIRECTORY:
            filelist_key = DatabaseType.FILELIST + key[1:]
            del self.db[filelist_key]

        p = "/".join(metadata["path"].split("/")[:-1])
        parent_filelist_key = keyify(DatabaseType.FILELIST, p)

        if parent_filelist_key in self.db:
            self.db[parent_filelist_key].remove(key)
            self.db[parent_filelist_key] = self.db[parent_filelist_key]

    def reset_keys(self):
        self._touched_keys = defaultdict(int)
        self._filelists = defaultdict(set)

    def reset_buckets(self):
        self._buckets = defaultdict(set)
        self._uncommitted_changes = 0

    def commit_keys(self):
        for key, modified in self._touched_keys.items():
            m = self.db[key]
            if m.get("modified", 0) < modified:
                m["modified"] = modified
                self.db[key] = m

        for key, hashes in self._filelists.items():
            self.db[key] |= hashes

    def commit_buckets(self):
        for bucket, hashes in self._buckets.items():
            key = "%s%s" % (DatabaseType.NEW_BUCKET, bucket)
            self.db[key] = self.db[key] | hashes

        self.reset_buckets()

    def add_hash_to_bucket(self, h):
        if not self._in_session:
            raise NotInSessionException()

        bucket = h[1 : 1 + BUCKET_SIZE]

        self._buckets[bucket].add(h)
        self._uncommitted_changes += 1

    def add_parent_path_touched(self, path, modified_time):
        split_path = [x for x in path.split("/") if x]
        if len(split_path) <= 1:
            return

        split_path.pop()
        while True:
            key = keyify(DatabaseType.DIRECTORY, "/".join(split_path))
            self._touched_keys[key] = max(self._touched_keys[key], modified_time)
            if not split_path:
                break
            split_path.pop()

    def add_hash_to_parent_dir(self, path, h):
        split_path = [x for x in path.split("/") if x]
        split_path.pop()

        parent_key = keyify(DatabaseType.FILELIST, "/".join(split_path))
        self._filelists[parent_key].add(h)

    def check_for_commit(self):
        if self._uncommitted_changes >= COMMIT_COUNTER:
            self.commit_session()
            self.reset_session()

    def _add_item(self, key_type, path, add_time, metadata=None):
        key = keyify(key_type, path)
        is_modified = False

        m = self.db.get(key)
        if not m:
            m = {"path": path, "date": add_time, "modified": add_time}
            is_modified = True

            if key_type == DatabaseType.DIRECTORY:
                filelist_key = keyify(DatabaseType.FILELIST, path)
                self.db[filelist_key] = set()

        if metadata and not compare_dicts(metadata, m):
            m.update(metadata)
            is_modified = True

        if "deleted" in m:
            del m["deleted"]
            is_modified = True

        if is_modified:
            self.add_parent_path_touched(path, add_time)
            m["modified"] = add_time

            self.add_hash_to_parent_dir(path, key)

            self.db[key] = m

        self.add_hash_to_bucket(key)

        self.check_for_commit()

    def add_file(self, path, size, add_time, metadata=None):
        assert self._in_session, "Not in session"

        metadata = metadata or {}
        metadata["size"] = size
        self._add_item(DatabaseType.FILE, path, add_time, metadata)

    def add_dir(self, path, add_time, metadata=None):
        assert self._in_session, "Not in session"

        metadata = metadata or {}
        self._add_item(DatabaseType.DIRECTORY, path, add_time, metadata)

    def last_modified(self, path):
        path = cleanup_path(path)

        key = keyify(DatabaseType.DIRECTORY, path)
        if key not in self.db:
            key = keyify(DatabaseType.FILE, path)
            if key not in self.db:
                raise PathNotFoundException()
        metadata = self.db[key]

        return metadata["modified"] if "modified" in metadata else metadata["date"]

    def update_metadata(self, db_type, path, metadata):
        assert self._in_session, "Not in session"

        key = keyify(db_type, path)
        m = self.db.get(key, None)
        if m is None:
            raise PathNotFoundException()

        modified = False
        for k, v in metadata.items():
            if m.get(k) != v:
                modified = True

        if modified:
            m.update(metadata)
            self.db[key] = m

            if "modified" not in metadata:
                self.add_parent_path_touched(path, self._current_time)

    def get_metadata(self, db_type, path):
        m = self.db.get(keyify(db_type, path), None)
        if m is None:
            raise PathNotFoundException()
        return m

    def _list_dir(self, parent_folder, path, depth, show_deleted):
        filelist_key = keyify(DatabaseType.FILELIST, path)

        for key in self.db[filelist_key]:
            item_type = key[0]
            metadata = self.db[key].copy()

            if not show_deleted and metadata.get("deleted", False):
                continue

            item_path = metadata["path"]
            item = Item(id=item_path.split("/")[-1], attributes=metadata, router=router)
            parent_folder.add_item(item)

            if item_type == DatabaseType.DIRECTORY:
                if depth:
                    item.initiate_nested_items()
                    self._list_dir(item, item_path, depth - 1, show_deleted)
                else:
                    item.expandable = True
                    self._trigger_event("list_route_needed", item=item, path=item_path)
            elif item_type == DatabaseType.FILE:
                item.readable = True
                self._trigger_event("file_route_needed", item=item, path=item_path)

            self._trigger_event("on_item", item=item, path=item_path)

        parent_folder.nested_items.sort(key=lambda x: x.id)

    def list_dir(self, path, depth=0, show_deleted=False):
        path = cleanup_path(path)

        key = keyify(DatabaseType.DIRECTORY, path)
        if key not in self.db:
            raise PathNotFoundException()
        metadata = self.db[key].copy()

        item_path = metadata["path"]
        item = Item(id=metadata["path"].split("/")[-1], attributes=metadata)

        if depth >= 0:
            item.initiate_nested_items()
            self._list_dir(item, path, depth, show_deleted)
        else:
            item.expandable = True
            self._trigger_event("list_route_needed", item=item, path=item_path)

        self._trigger_event("on_item", item=item, path=item_path)

        return item

    def list_file(self, path):
        path = cleanup_path(path)

        key = keyify(DatabaseType.FILE, path)
        if key not in self.db:
            raise PathNotFoundException()
        metadata = self.db[key].copy()

        item_path = metadata["path"]
        item = Item(id=metadata["path"].split("/")[-1], attributes=metadata)

        item.readable = True
        self._trigger_event("file_route_needed", item=item, path=item_path)
        self._trigger_event("on_item", item=item, path=item_path)

        return item



import json
import logging

from django.http import Http404
from rest_framework.response import Response
from rest_framework.views import APIView
from thomas import Item
from unplugged import CascadingPermission

from ...plugins import SearchQuery
from ...utils import hash_string
from .models import ListingCache

logger = logging.getLogger(__name__)


class RemoteSearcherFiltersView(APIView):
    service = None

    permission_classes = (CascadingPermission,)

    def get(self, request):
        return Response(
            [
                {
                    "name": searcher["searcher"].name,
                    "filters": searcher["searcher"].filters.serialize(),
                }
                for searcher in self.service.config["searchers"]
            ]
        )


class RemoteSearcherView(APIView):
    service = None

    permission_classes = (CascadingPermission,)

    def get_item(self, searcher_name, search_token, path, sub_path):
        try:
            listing_cache = ListingCache.objects.get(
                app=self.service.name,
                searcher_name=searcher_name,
                search_token=search_token,
                path=path,
            )
        except ListingCache.DoesNotExist:
            logger.warning(
                f"Unable to find cache for search_token:{search_token} path:{path}"
            )
            raise Http404

        return Item.unserialize(listing_cache.listing).get_item_from_path(sub_path)

    def get_searcher(self, searcher_name):
        for searcher in self.service.config["searchers"]:
            searcher = searcher["searcher"]
            if searcher.name == searcher_name:
                return searcher
        else:
            logger.warning(f"Unknown searcher:{searcher_name}")
            raise Http404

    def get(self, request, searcher_name, search_token="", path=""):
        searcher = self.get_searcher(searcher_name)

        path = path.strip("/")
        sub_path = ""

        if search_token:
            sub_path = request.GET.get("sub_path").strip("/")
            item = self.get_item(searcher_name, search_token, path, sub_path)
            if not item:
                logger.warning(
                    f"Unknown item search_token:{search_token} path:{path} sub_path:{sub_path}"
                )
                raise Http404
        else:
            query = dict(request.GET)
            search_token = hash_string(json.dumps(sorted(query.items())))
            search_query = SearchQuery(searcher.filters, query)

            logger.info(
                f"Got request to search for search_token:{search_token} query:{query!r}"
            )
            item = searcher.get_item(search_token, search_query)

        full_path = (f"{path}/{sub_path}").strip("/")
        logger.debug(
            f"Listing search_token:{search_token} path:{full_path} item:{item!r}"
        )
        item.list()
        ListingCache.objects.update_or_create(
            app=self.service.name,
            searcher_name=searcher_name,
            search_token=search_token,
            path=full_path,
            defaults={"listing": item.serialize(include_routes=True)},
        )

        availability_plugin = self.service.config.get("availability")
        if availability_plugin:
            logger.info("Found availability plugin, populating item.")
            availability_plugin.populate_available(item)

        return Response(
            {
                "item": item.serialize(),
                "path": full_path,
                "search_token": search_token,
                "searcher_name": searcher_name,
            }
        )

    def post(self, request, searcher_name, search_token="", path=""):
        path = path.strip("/")
        sub_path = request.GET.get("sub_path").strip("/")
        if not search_token:
            logger.warning("Search token missing from stream attempt")
            raise Http404

        item = self.get_item(searcher_name, search_token, path, sub_path)
        stream = item.stream()
        if not stream:
            logger.warning("Failed to return a stream")
            raise Http404

        return Response(stream.serialize(request))



import json
import logging
import time
from abc import abstractmethod, abstractproperty

from unplugged import PluginBase, threadify

from twisted.internet import reactor

from ..utils import hash_string

logger = logging.getLogger(__name__)


class SearchQuery(dict):
    def __init__(self, searcher_filter, query):
        self.order_by = []

        for k, v in query.items():
            if k == "o":
                if isinstance(v, list):
                    self.order_by.extend(v)
                else:
                    self.order_by.append(v)
            else:
                if k not in searcher_filter.fields:
                    continue

                if k in searcher_filter.choices:
                    if not isinstance(v, list):
                        v = [v]

                    self.setdefault(k, [])
                    for val in v:
                        if val not in searcher_filter.choices[k]:
                            continue

                        self[k].append(val)
                else:
                    self[k] = v

    def get_hash(self):
        hash_string(json.dumps(sorted(self.items())))


class SearcherFilter:
    def __init__(self, fields):
        self.fields = fields
        self.choices = {}
        self.order_by = []

    def set_choices(self, field, choices):
        if field not in self.fields:
            logger.warning(f"Trying to set choices for unknown field {field}")

        self.choices[field] = choices

    def add_order_by(self, field):
        self.order_by.append(field)

    def serialize(self):
        if "o" in self.fields:
            logger.warning(
                'field "o" is added as search field. As it is used as order field, it will cause conflict.'
            )

        return {
            "fields": self.fields,
            "choices": self.choices,
            "order_by": self.order_by,
        }

    @classmethod
    def unserialize(cls, data):
        obj = cls(data["fields"])
        obj.choices = data["choices"]
        obj.order_by = data["order_by"]
        return obj

    def merge(self, searcher_filter):
        fields = list(set(self.fields + searcher_filter.fields))
        sf = SearcherFilter(fields)

        for key in set(searcher_filter.choices.keys() + self.choices.keys()):
            choices = list(
                set(searcher_filter.choices.get(key, []) + self.choices.get(key, []))
            )
            sf.set_choices(key, choices)

        for order_by in set(searcher_filter.order_by) & set(self.order_by):
            sf.add_order_by(order_by)

        return sf


class SearcherPlugin(PluginBase):
    plugin_type = "searcher"

    @abstractmethod
    def get_item(query_hash, search_query):
        """
        Return an item for this search query
        """
        raise NotImplementedError

    @abstractproperty
    def filters(self):
        """
        Returns instance of SearcherFilter.
        """
        raise NotImplementedError


class SearcherPluginManager:
    @staticmethod
    def get_item_multiple(
        plugins, query_hash, search_query
    ):  # get an item for all paths and merge into one
        item = None
        for plugin in plugins:
            plugin_item = plugin.get_item(query_hash, search_query)
            if item:
                item.merge(plugin_item)
            else:
                item = plugin_item

        item["modified"] = int(time.time())
        return item

    @staticmethod
    def get_item(plugin, query_hash, search_query):
        return SearcherPluginManager.get_item_multiple(
            [plugin], query_hash, search_query
        )

    @staticmethod
    def filters_multiple(plugins):
        threads = []
        for plugin in plugins:

            def get_filters(plugin):
                return plugin.filters

            threads.append((plugin, threadify(get_filters, cache_result=True)(plugin)))

        retval = None
        for plugin, thread in threads:
            filters = thread()
            if not filters:
                continue

            if retval is None:
                retval = filters
            else:
                retval.merge(filters)

        return retval



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.metadata.embedded"
    verbose_name = "Embedded Metadata"
    label = "metadata_embedded"

    def ready(self):
        from .handler import EmbeddedMetadataHandlerPlugin  # NOQA



import logging
import threading
import time
from functools import partial

from django.conf import settings
from unplugged import ServicePlugin, command, threadify
from unplugged.models import Schedule

from .listingbuilder import ListingBuilder
from .models import ListingItem

logger = logging.getLogger(__name__)


class BaseListingService(ServicePlugin):
    _do_unload = False
    can_automatically_rebuild = True
    automatic_rebuild_timer = 15

    def __init__(self, config):
        self.config = config
        self.listing_builder = ListingBuilder(self)
        self.automatic_rebuild_lock = threading.Lock()

        if self.can_automatically_rebuild:
            # TODO: should run after bootstrap
            threadify(self.schedule_automatic_rebuild, delay=3)()
            threadify(self.rebuild_listings, delay=30)()

    def get_section(self, section_name):
        for section in self.config.get("sections", []):
            if section["name"] == section_name:
                return section.copy()

    @command(
        name="clear",
        display_name="Clear",
        description="Clear all cached listings and rebuild them on access",
    )
    def command_clear(self):
        logger.info("Clearing cached listings")
        ListingItem.objects.filter(app=self.name).delete()

    def schedule_automatic_rebuild(self):
        Schedule.objects.ensure_schedule(
            Schedule.METHOD_INTERVAL,
            f"minutes={self.automatic_rebuild_timer}",
            "rebuild_listings",
            {},
            self,
            "automatic_rebuild_listing",
        )

    @command(
        name="rebuild_listings",
        display_name="Rebuild listings",
        description="Try to rebuild all the listings that allows for automatic rebuild",
    )
    def rebuild_listings(self):
        logger.debug(f"We got call for automatic rebuild for {self.name}")
        if settings.DEBUG_CLEAR_CACHE:
            logger.debug("Skipping automatic rebuild due to clear cache debug option")
            return

        if self.automatic_rebuild_lock.locked():
            logger.debug("Automatic rebuild already in progress, bailing")
            return

        with self.automatic_rebuild_lock:
            for section in self.config.get("sections", []):
                logger.debug(
                    f"Checking if we can rebuild service:{self.name} section:{section['name']}"
                )
                if self._do_unload:
                    logger.debug(
                        f"Not rebuilding because we are unloading service:{self.name} section:{section['name']}"
                    )
                    return

                if not section.get("rebuild_automatically"):
                    logger.debug(
                        f"This one does not automatically rebuild at all service:{self.name} section:{section['name']}"
                    )
                    continue

                config = self.get_path_config(section["name"])
                if not config:
                    logger.debug(
                        f"No config found service:{self.name} section:{section['name']}"
                    )
                    continue

                logger.debug(
                    f"Automatic rebuild for service:{self.name} section:{section['name']}"
                )
                self.rebuild_listing(config, section["name"])

    # TODO:
    # if queued, do not build
    # if queued the last 10 seconds, do not build
    def rebuild_listing(self, config, path, delay=False):
        threadify(self.listing_builder.get_listing, delay=delay and 3.0 or 0)(
            config, path, use_background_recheck=False
        )

    def unload(self):
        self._do_unload = True

        super().unload()

    def get_path_config(self, path):
        """
        Returns config for a given path, if None is returned then the lister will try to see
        if the listingitem already exists.
        """
        path = path.split("/")
        section = path.pop(0)
        depth = len(path)
        path = "/".join(path)

        logger.debug(f"Looking for path:{path} in section:{section} with depth:{depth}")

        config = self.get_section(section)
        if not config:
            logger.info("Section config returned None")
            return None

        max_depth = 0
        for level_config in config["levels"]:
            max_depth = max(max_depth, level_config["depth"])
            if level_config["depth"] == depth:
                config["level"] = level_config.copy()

        config["parent_level_config"] = None
        config["path"] = path
        config["section"] = section
        config["edge_type"] = (
            config.get("level", {}).get("depth", -1) == max_depth and "file" or "folder"
        )

        config["levels"] = sorted(config["levels"], key=lambda x: x["depth"])
        for i, level_config in enumerate(config["levels"]):
            if level_config["depth"] == depth:
                config["current_level"] = i
                break
            else:
                config["parent_level_config"] = level_config

        return config

    def get_nearest_config(self, path):
        """
        Finds the nearest parent config and returns it including parent_path
        """
        config = None
        parent_path = path
        while parent_path and (not config or config.get("level") is None):
            parent_path = parent_path.rsplit("/", 1)[0]
            config = self.get_path_config(parent_path)

            if "/" not in parent_path:
                break

        return config, parent_path



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "tridentstream.services.metadata"
    verbose_name = "Metadata Service"
    label = "services_metadata"

    def ready(self):
        from .handler import MetadataServicePlugin  # NOQA
